__doc__ = """
Transverse muscle implementation
"""

from typing import Union

import numpy as np
from numba import njit

from coomm.actuations.muscles.muscle import MuscleForce


class TransverseMuscle(MuscleForce):
    """TransverseMuscle."""

    def __init__(
        self,
        rest_muscle_area: np.ndarray,
        max_muscle_stress: Union[float, np.ndarray],
        **kwargs
    ):
        """__init__.

        Parameters
        ----------
        rest_muscle_area : np.ndarray
            shape: (n_element)
        max_muscle_stress : Union[float, np.ndarray]
            shape: (n_element)
        """
        n_elem = rest_muscle_area.shape[0]
        super().__init__(
            ratio_muscle_position=np.zeros((3, n_elem)),
            rest_muscle_area=rest_muscle_area,
            max_muscle_stress=-max_muscle_stress,
            type_name="TM",
            **kwargs
        )

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_length(muscle_length: np.ndarray, muscle_strain: np.ndarray):
        """calculate_muscle_length.

        Parameters
        ----------
        muscle_length : np.ndarray
            shape: (n_element)
        muscle_strain : np.ndarray
            shape: (3, n_element)
        """
        blocksize = muscle_length.shape[0]
        for i in range(blocksize):
            # fmt: off
            muscle_length[i] = (
                1 / (muscle_strain[0, i] ** 2 +  \
                     muscle_strain[1, i] ** 2 +  \
                     muscle_strain[2, i] ** 2) ** 0.25 \
            )
            # fmt: on



"""
Created on Dec. 23, 2021
@author: Heng-Sheng (Hanson) Chang
"""

from coomm.povray import POVRAYBase
from coomm.povray.muscles.muscle import POVRAYRingMuscle

class POVRAYTransverseMuscle(POVRAYBase, POVRAYRingMuscle):
    def __init__(self, **kwargs):
        POVRAYBase.__init__(self, **kwargs)
        POVRAYRingMuscle.__init__(self, **kwargs)
        self.color_string = self.to_color_string(self.muscle_color)
        self.muscle_label = "// transverse muscle data\n"



__doc__ = """
Frame tool implementations
"""
import numpy as np
import matplotlib.colors as mcolors

# TODO: maybe combine default plotting parameters?
default_colors = mcolors.TABLEAU_COLORS
# default_colors['tab:blue']    # muted blue
# default_colors['tab:orange']  # safety orange
# default_colors['tab:green']   # cooked asparagus green
# default_colors['tab:red']     # brick red
# default_colors['tab:purple']  # muted purple
# default_colors['tab:brown']   # chestnut brown
# default_colors['tab:pink']    # raspberry yogurt pink
# default_colors['tab:gray']    # middle gray
# default_colors['tab:olive']   # curry yellow-green
# default_colors['tab:cyan']    # blue-teal

base_colors = mcolors.BASE_COLORS
# base_colors['b']              # blue
# base_colors['g']              # green
# base_colors['r']              # red
# base_colors['c']              # cyan
# base_colors['m']              # magenta
# base_colors['y']              # yellow
# base_colors['k']              # black
# base_colors['w']              # white
    
default_label_fontsize = 15
paper_label_fontsize = 48
paper_linewidth = 5

# @njit(cache=True)
# def process_position(position, offset, rotation):
#     blocksize = position.shape[1]
#     output_position = np.zeros((3, blocksize))
#     for n in range(blocksize):
#         for i in range(3):
#             for j in range(3):
#                 output_position[i, n] += (
#                     rotation[i, j] * (position[j, n] - offset[j])
#                 )
#     return output_position

# @njit(cache=True)
# def process_director(director, rotation):
#     blocksize = director.shape[2]
#     output_director = np.zeros((3, 3, blocksize))
#     for n in range(blocksize):
#         for i in range(3):
#             for j in range(3):
#                 for k in range(3):
#                     output_director[i, j, n] += (
#                         rotation[i, k] * director[k, j, n]
#                     )
#     return output_director

def change_box_to_arrow_axes(
    fig, ax, linewidth=1.0, overhang=0.0,
    xaxis_ypos=0, yaxis_xpos=0,
    x_offset=[0, 0], y_offset=[0, 0], # TODO: Never use mutable (list) object for default
    color='black'
):
    """change_box_to_arrow_axes.

    Parameters
    ----------
    fig :
    ax :
    linewidth :
    overhang :
    xaxis_ypos :
    yaxis_xpos :
    x_offset :
    y_offset :
    color :
    """
    for spine in ax.spines.values():
        spine.set_visible(False)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    # get width and height of axes object to compute
    # matching arrowhead length and width
    dps = fig.dpi_scale_trans.inverted()
    bbox = ax.get_window_extent().transformed(dps)
    width, height = bbox.width, bbox.height

    # manual arrowhead width and length
    hw = 1./40.*(ymax-ymin)
    hl = 1./40.*(xmax-xmin)
    lw = linewidth # axis line width
    ohg = overhang # arrow overhang

    # compute matching arrowhead length and width
    yhw = hw/(ymax-ymin)*(xmax-xmin)* height/width
    yhl = hl/(xmax-xmin)*(ymax-ymin)* width/height

    # draw x and y axis
    start_x = xmin + x_offset[0]
    dx = xmax + x_offset[1] - start_x
    ax.arrow(start_x, xaxis_ypos, dx, 0, fc=color, ec=color, lw=lw, 
                head_width=hw, head_length=hl, overhang=ohg, 
                length_includes_head= True, clip_on=False) 

    start_y = ymin+y_offset[0]
    dy = ymax + y_offset[1] - start_y
    ax.arrow(yaxis_xpos, start_y, 0, dy, fc=color, ec=color, lw=lw, 
                head_width=yhw, head_length=yhl, overhang=ohg, 
                length_includes_head= True, clip_on=False)
    return ax

def change_box_to_only_y_arrow_ax(
    fig, ax, linewidth=1.0, overhang=0.0,
    yaxis_xpos=0, y_offset=[0, 0], # TODO: Never use mutable (list) object for default
    color='black'
):
    """change_box_to_only_y_arrow_ax.

    Parameters
    ----------
    fig :
    ax :
    linewidth :
    overhang :
    yaxis_xpos :
    y_offset :
    color :
    """
    for spine in ax.spines.values():
        spine.set_visible(False)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    # get width and height of axes object to compute
    # matching arrowhead length and width
    dps = fig.dpi_scale_trans.inverted()
    bbox = ax.get_window_extent().transformed(dps)
    width, height = bbox.width, bbox.height

    # manual arrowhead width and length
    hw = 1./40.*(ymax-ymin)
    hl = 1./40.*(xmax-xmin)
    lw = linewidth # axis line width
    ohg = overhang # arrow overhang

    # compute matching arrowhead length and width
    yhw = hw/(ymax-ymin)*(xmax-xmin)* height/width
    yhl = hl/(xmax-xmin)*(ymax-ymin)* width/height

    # draw y axis
    start_y = ymin+y_offset[0]
    dy = ymax + y_offset[1] - start_y
    ax.arrow(yaxis_xpos, start_y, 0, dy, fc=color, ec=color, lw=lw, 
                head_width=yhw, head_length=yhl, overhang=ohg, 
                length_includes_head=True, clip_on=False)
    return ax

def change_box_to_only_x_line_ax(
    fig, ax, linewidth=1.0,
    xaxis_ypos=0, x_offset=[0, 0], # TODO: Never use mutable (list) object for default
    color='black'
):
    """change_box_to_only_x_line_ax.

    Parameters
    ----------
    fig :
    ax :
    linewidth :
    xaxis_ypos :
    x_offset :
    color :
    """
    for spine in ax.spines.values():
        spine.set_visible(False)

    xmin, xmax = ax.get_xlim()

    # draw x axis
    start_x = xmin + x_offset[0]
    dx = xmax + x_offset[1] - start_x
    ax.plot(
        [start_x, start_x+dx],
        [xaxis_ypos, xaxis_ypos],
        linewidth=linewidth,
        color=color
    )
    return ax

def add_y_ticks(ax, yticks, ticks_xpos, length, linewidth, color='black'):
    """add_y_ticks.

    Parameters
    ----------
    ax :
    yticks :
    ticks_xpos :
    length :
    linewidth :
    color :
    """
    ax.set_yticks(yticks)
    for ytick in yticks:
        ax.plot(
            [ticks_xpos, ticks_xpos+length],
            [ytick, ytick],
            linewidth=linewidth,
            color=color
        )



__doc__ = """
Muscle base class implementation.
"""

from typing import Union, Iterable, Dict

from collections import defaultdict
import numpy as np
from numba import njit

import elastica
from elastica._linalg import _batch_cross
from elastica._calculus import quadrature_kernel
from elastica.external_forces import inplace_addition
from coomm._rod_tool import average2D, difference2D, sigma_to_shear

from coomm.actuations.actuation import (
    _force_induced_couple,
    _internal_to_external_load,
    ContinuousActuation,
    ApplyActuations,
)


@njit(cache=True)
def force_length_weight_guassian(
    muscle_length: np.ndarray, sigma: float = 0.25
) -> np.ndarray:
    force_weight = np.exp(-0.5 * ((muscle_length - 1) / sigma) ** 2)
    return force_weight


# force-length curve (x) = 3.06 x^3 - 13.64 x^2 + 18.01 x - 6.44
@njit(cache=True)
def force_length_weight_poly(
    muscle_length: np.ndarray,
    f_l_coefficients: np.ndarray = np.array([-6.44, 18.01, -13.64, 3.06]),
) -> np.ndarray:  # FIXME: This work?? with numba??? In any way, parameters should be immutable
    degree = f_l_coefficients.shape[0]

    blocksize = muscle_length.shape[0]
    force_weight = np.zeros(blocksize)
    for i in range(blocksize):
        for power in range(degree):
            force_weight[i] += f_l_coefficients[power] * (muscle_length[i] ** power)
        force_weight[i] = (
            0 if (force_weight[i] < 0) or (muscle_length[i] > 2) else force_weight[i]
        )
    return force_weight


class MuscleInfo:
    # TODO: Maybe try to implement this class as @dataclass
    """MuscleInfo.
    Data class to store muscle's meta data
    """

    def __init__(self, type_name: str="muscle", index: int=0, **kwargs) -> None:
        """Muscle information constructor

        Parameters
        ----------
        type_name : str
            Name of the muscle type
        index : int
            Muscle index
        """
        super().__init__(**kwargs)
        self.type_name = type_name
        self.index = index

    def __str__(self) -> str:
        return f"{self.index}_{self.type_name}"


class Muscle(MuscleInfo, ContinuousActuation):
    """Muscle base class"""

    def __init__(
        self,
        ratio_muscle_position: np.ndarray,
        rest_muscle_area: np.ndarray,
        index: int = 0,
        **kwargs,
    ) -> None:
        """__init__.

        Parameters
        ----------
        ratio_muscle_position : np.ndarray
            shape: (3, n_element)
        rest_muscle_area : np.ndarray
            shape: (n_element)
        index : int
        """
        super().__init__(
            n_elements=rest_muscle_area.shape[0],
            index=index,
            **kwargs,
        )

        self.s = np.linspace(0, 1, self.n_elements + 1)
        self.muscle_normalized_length = np.zeros(self.n_elements)
        self.muscle_rest_length = np.ones(self.n_elements)
        self.muscle_length = np.zeros(self.n_elements)
        self.muscle_tangent = np.zeros((3, self.n_elements))
        self.muscle_strain = np.zeros((3, self.n_elements))
        self.muscle_position = np.zeros((3, self.n_elements))
        self.ratio_muscle_position = ratio_muscle_position.copy()
        self.rest_muscle_area = rest_muscle_area.copy()
        self.muscle_area = self.rest_muscle_area.copy()

    def __call__(self, system: elastica.rod.RodBase) -> None:
        """__call__.

        Parameters
        ----------
        system : elastica.rod.RodBase
        """
        self.calculate_muscle_area(
            self.rest_muscle_area, self.muscle_area, system.dilatation
        )
        self.calculate_muscle_position(
            self.muscle_position, system.radius, self.ratio_muscle_position
        )
        self.calculate_muscle_strain(
            self.muscle_strain,
            self.muscle_position,
            system.sigma,
            system.kappa,
            system.rest_voronoi_lengths,
            system.voronoi_dilatation,
        )
        self.calculate_muscle_tangent(self.muscle_tangent, self.muscle_strain)

    def set_current_length_as_rest_length(self, system: elastica.rod.RodBase) -> None:
        """set_current_length_as_rest_length.

        Parameters
        ----------
        system : elastica.rod.RodBase
        """
        self.__call__(system)
        self.calculate_muscle_length(self.muscle_length, self.muscle_strain)
        self.muscle_rest_length[:] = self.muscle_length

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_area(rest_muscle_area, muscle_area, dilatation):
        muscle_area[:] = rest_muscle_area / dilatation

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_position(muscle_position, radius, ratio_muscle_position):
        muscle_position[:, :] = radius * ratio_muscle_position

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_strain(
        muscle_strain,
        off_center_displacement,
        sigma,
        kappa,
        rest_voronoi_lengths,
        voronoi_dilatation,
    ):
        shear = sigma_to_shear(sigma)
        muscle_position_derivative = difference2D(off_center_displacement) / (
            rest_voronoi_lengths * voronoi_dilatation
        )
        muscle_strain[:, :] = shear + quadrature_kernel(
            _batch_cross(kappa, average2D(off_center_displacement))
            + muscle_position_derivative
        )

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_tangent(muscle_tangent, muscle_strain):
        blocksize = muscle_strain.shape[1]
        for i in range(blocksize):
            muscle_tangent[:, i] = muscle_strain[:, i] / np.sqrt(
                muscle_strain[0, i] ** 2
                + muscle_strain[1, i] ** 2
                + muscle_strain[2, i] ** 2
            )

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_length(muscle_length, muscle_strain):
        blocksize = muscle_length.shape[0]
        for i in range(blocksize):
            muscle_length[i] = np.sqrt(
                muscle_strain[0, i] ** 2
                + muscle_strain[1, i] ** 2
                + muscle_strain[2, i] ** 2
            )

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_normalized_length(
        muscle_normalized_length, muscle_length, muscle_rest_length
    ):
        muscle_normalized_length[:] = muscle_length / muscle_rest_length


class MuscleForce(Muscle):
    """MuscleForce"""

    def __init__(
        self,
        ratio_muscle_position: np.ndarray,
        rest_muscle_area: np.ndarray,
        max_muscle_stress: Union[float, np.ndarray],
        **kwargs,
    ):
        """
        Muscle force class implementation

        Parameters
        ----------
        ratio_muscle_position : np.ndarray
            shape: (3, n_element)
        rest_muscle_area : np.ndarray
            shape: (n_element)
        max_muscle_stress : Union[float, np.array]
            shape: (n_element)
        """
        super().__init__(
            ratio_muscle_position=ratio_muscle_position,
            rest_muscle_area=rest_muscle_area,
            muscle_type="muscle_force",
            **kwargs,
        )
        self.activation = np.zeros(self.n_elements)
        self.s_activation = (self.s[:-1] + self.s[1:]) / 2
        if isinstance(max_muscle_stress, float):
            self.max_muscle_stress = max_muscle_stress
        elif isinstance(max_muscle_stress, np.ndarray):
            self.max_muscle_stress = max_muscle_stress.copy()
        else:
            raise TypeError(
                f"{max_muscle_stress=} must be either float or np.ndarray. "
            )
        self.muscle_force = np.zeros(self.n_elements)
        self.s_force = 0.5 * (self.s[:-1] + self.s[1:])
        self.force_length_weight = kwargs.get("force_length_weight", np.ones_like)

    def __call__(self, system: elastica.rod.RodBase):
        """__call__.

        Parameters
        ----------
        system : elastica.rod.RodBase
        """
        super().__call__(system)
        self.calculate_muscle_length(self.muscle_length, self.muscle_strain)
        self.calculate_muscle_normalized_length(
            self.muscle_normalized_length, self.muscle_length, self.muscle_rest_length
        )
        self.calculate_muscle_force(
            self.muscle_force,
            self.get_activation(),
            self.max_muscle_stress,
            self.force_length_weight(self.muscle_normalized_length),
            self.muscle_area,
        )
        self.calculate_force_and_couple(
            self.muscle_force,
            self.muscle_tangent,
            self.muscle_position,
            self.internal_force,
            self.internal_couple,
            self.external_force,
            self.external_couple,
            system.director_collection,
            system.kappa,
            system.tangents,
            system.rest_lengths,
            system.rest_voronoi_lengths,
            system.dilatation,
            system.voronoi_dilatation,
        )

    @staticmethod
    @njit(cache=True)
    def calculate_muscle_force(
        muscle_force, muscle_activation, max_muscle_stress, weight, muscle_area
    ):
        muscle_force[:] = (muscle_activation * max_muscle_stress * weight) * muscle_area

    @staticmethod
    @njit(cache=True)
    def calculate_force_and_couple(
        muscle_force,
        muscle_tangent,
        muscle_position,
        internal_force,
        internal_couple,
        external_force,
        external_couple,
        director_collection,
        kappa,
        tangents,
        rest_lengths,
        rest_voronoi_lengths,
        dilatation,
        voronoi_dilatation,
    ):
        internal_force[:, :] = muscle_force * muscle_tangent
        _force_induced_couple(internal_force, muscle_position, internal_couple)
        _internal_to_external_load(
            director_collection,
            kappa,
            tangents,
            rest_lengths,
            rest_voronoi_lengths,
            dilatation,
            voronoi_dilatation,
            internal_force,
            internal_couple,
            external_force,
            external_couple,
        )

    def apply_activation(self, activation: Union[float, np.ndarray]):
        """apply_activation.

        Parameters
        ----------
        activation : Union[float, np.ndarray]
            If array of activation is given, the shape of activation is expected to
            match the shape of muscle_activation.
        """
        self.set_activation(self.activation, activation)

    @staticmethod
    @njit(cache=True)
    def set_activation(muscle_activation, activation):
        muscle_activation[:] = activation

    def get_activation(self) -> Union[float, np.ndarray]:
        """
        activation getter

        Returns
        -------
        activation: Union[float, np.ndarray]
        """
        return self.activation


class MuscleGroup(MuscleInfo, ContinuousActuation):
    """MuscleGroup.
    Group of muscle. Provides convinience tools to operate group-activation.
    """

    def __init__(
        self,
        muscles: Iterable[Muscle],
        type_name: str = "muscle_group",
        index: int = 0,
        **kwargs,
    ):
        """__init__.

        Parameters
        ----------
        muscles : Iterable[Muscle]
        """
        super().__init__(
            n_elements=muscles[0].n_elements, type_name=type_name, index=index, **kwargs
        )

        self.muscles = muscles
        for m, muscle in enumerate(self.muscles):
            muscle.index = m
        self.activation = np.zeros(self.muscles[0].activation.shape)
        self.s_activation = self.muscles[0].s_activation.copy()

    def __call__(self, system: elastica.rod.RodBase):
        """__call__.

        Parameters
        ----------
        system : elastica.rod.RodBase
        """
        self.reset_actuation()
        for muscle in self.muscles:
            muscle(system)
            inplace_addition(self.internal_force, muscle.internal_force)
            inplace_addition(self.external_force, muscle.external_force)
            inplace_addition(self.internal_couple, muscle.internal_couple)
            inplace_addition(self.external_couple, muscle.external_couple)

    def set_current_length_as_rest_length(self, system: elastica.rod.RodBase):
        """set_current_length_as_rest_length.

        Parameters
        ----------
        system : elastica.rod.RodBase
        """
        for muscle in self.muscles:
            muscle.set_current_length_as_rest_length(system)

    def apply_activation(self, activation: Union[float, np.ndarray]):
        """apply_activation.

        MuscleGroup apply activation

        Parameters
        ----------
        activation : Union[float, np.ndarray]
            If array of activation is given, the shape of activation is expected to
            match the shape of muscle_activation.
        """
        self.set_activation(self.activation, activation)
        for muscle in self.muscles:
            muscle.set_activation(muscle.activation, self.activation)

    @staticmethod
    @njit(cache=True)
    def set_activation(muscle_activation, activation):
        muscle_activation[:] = activation

    def get_activation(self) -> Union[float, np.ndarray]:
        """
        activation getter

        Returns
        -------
        activation: Union[float, np.ndarray]
        """
        return self.activation


class ApplyMuscles(ApplyActuations):
    """ApplyMuscles."""

    def __init__(
        self, muscles: Iterable[Muscle], step_skip: int, callback_params_list: list
    ):
        """__init__.

        Parameters
        ----------
        muscles : Iterable[Muscle]
        step_skip : int
        callback_params_list : list
        """
        super().__init__(muscles, step_skip, callback_params_list)
        for m, muscle in enumerate(muscles):
            muscle.index = m

    def callback_func(
        self, muscles: Iterable[Muscle], callback_params_list: Iterable[Dict]
    ):
        """callback_func.

        Parameters
        ----------
        muscles : Iterable[Muscle]
        callback_params_list : Iterable[Dict]
        """
        for muscle, callback_params in zip(muscles, callback_params_list):
            callback_params["muscle_info"].append(str(muscle))
            callback_params["s_activation"].append(muscle.s_activation.copy())
            callback_params["activation"].append(muscle.activation.copy())
            callback_params["muscle_length"].append(muscle.muscle_length.copy())
            callback_params["muscle_normalized_length"].append(
                muscle.muscle_normalized_length.copy()
            )
            callback_params["force_length_weight"].append(
                muscle.force_length_weight(muscle.muscle_normalized_length).copy()
            )
            callback_params["muscle_position"].append(muscle.muscle_position.copy())
            callback_params["internal_force"].append(muscle.internal_force.copy())
            callback_params["internal_couple"].append(muscle.internal_couple.copy())
            callback_params["external_force"].append(muscle.external_force.copy())
            callback_params["external_couple"].append(muscle.external_couple.copy())


class ApplyMuscleGroups(ApplyMuscles):
    """ApplyMuscleGroups."""

    def __init__(
        self, muscle_groups: MuscleGroup, step_skip: int, callback_params_list: list
    ):
        """__init__.

        Parameters
        ----------
        muscle_groups : MuscleGroup
        step_skip : int
        callback_params_list : list
        """
        super().__init__(muscle_groups, step_skip, callback_params_list)
        for muscle_group, callback_params in zip(
            muscle_groups, self.callback_params_list
        ):
            callback_params["muscles"] = [
                defaultdict(list) for _ in muscle_group.muscles
            ]

    def callback_func(
        self, muscle_groups: MuscleGroup, callback_params_list: Iterable[Dict]
    ):
        """callback_func.

        Parameters
        ----------
        muscle_groups : MuscleGroup, Iterable[Muscle]
        callback_params_list : Iterable[Dict]
        """
        for muscle_group, callback_params in zip(muscle_groups, callback_params_list):
            callback_params["muscle_group_info"].append(str(muscle_group))
            callback_params["s_activation"].append(muscle_group.s_activation.copy())
            callback_params["activation"].append(muscle_group.activation.copy())
            callback_params["internal_force"].append(muscle_group.internal_force.copy())
            callback_params["internal_couple"].append(
                muscle_group.internal_couple.copy()
            )
            callback_params["external_force"].append(muscle_group.external_force.copy())
            callback_params["external_couple"].append(
                muscle_group.external_couple.copy()
            )
            ApplyMuscles.callback_func(
                self, muscle_group.muscles, callback_params["muscles"]
            )



__doc__ = """
Collection of rod data-processing kernels.
"""

import numpy as np
from numba import njit

from elastica._linalg import _batch_matvec, _batch_norm
from elastica._calculus import _difference, _average
from elastica._rotations import _get_rotation_matrix, _inv_rotate

@njit(cache=True)
def inverse(matrix_collection):
    output_matrix_collection = np.empty(matrix_collection.shape)
    for n in range(output_matrix_collection.shape[2]):
        output_matrix_collection[:, :, n] = (
            np.linalg.inv(matrix_collection[:, :, n])
        )
    return output_matrix_collection

@njit(cache=True)
def _lab_to_material(directors, lab_vectors):
    return _batch_matvec(directors, lab_vectors)

@njit(cache=True)
def _material_to_lab(directors, material_vectors):
    blocksize = material_vectors.shape[1]
    lab_vectors = np.zeros((3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            for j in range(3):
                lab_vectors[i, n] += (
                    directors[j, i, n] * material_vectors[j, n]
                )
    return lab_vectors

@njit(cache=True)
def average1D(vector_collection):
    blocksize = vector_collection.shape[0]-1
    output_vector = np.zeros(blocksize)
    for n in range(blocksize):
        output_vector[n] = (vector_collection[n]+vector_collection[n+1])/2
    return output_vector

@njit(cache=True)
def average2D(vector_collection):
    blocksize = vector_collection.shape[1]-1
    output_vector = np.zeros((3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            output_vector[i, n] = (
                (vector_collection[i, n]+vector_collection[i, n+1])/2
            )
    return output_vector

@njit(cache=True)
def difference2D(vector_collection):
    blocksize = vector_collection.shape[1]-1
    output_vector = np.zeros((3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            output_vector[i, n] = (
                vector_collection[i, n+1]-vector_collection[i, n]
            )
    return output_vector

@njit(cache=True)
def calculate_dilatation(sigma):
    shear = sigma_to_shear(sigma)
    dilatation = _batch_norm(shear)
    voronoi_dilatation = (dilatation[:-1] + dilatation[1:])/2
    return dilatation, voronoi_dilatation

# @njit(cache=True)
# def calculate_length(position_collection):
#     length = 0
#     for n in range(position_collection.shape[1]-1):
#         length += np.sqrt(
#             (position_collection[0, n+1] - position_collection[0, n])**2 +
#             (position_collection[1, n+1] - position_collection[1, n])**2 +
#             (position_collection[2, n+1] - position_collection[2, n])**2
#         )
#     return length

@njit(cache=True)
def calculate_distance_to_a_point(position_collection, point_position):
    blocksize = position_collection.shape[1]
    distance_collection = np.zeros(blocksize)
    for n in range(blocksize):
        distance_collection[n] = (
            (position_collection[0, n]-point_position[0])**2 +
            (position_collection[1, n]-point_position[1])**2 +
            (position_collection[2, n]-point_position[2])**2
        )**0.5
    return distance_collection

@njit(cache=True)
def calculate_distance(position_collection_1, position_collection_2):
    blocksize = position_collection_1.shape[1]
    distance_collection = np.zeros(blocksize)
    for n in range(blocksize):
        distance_collection[n] = (
            (position_collection_1[0, n]-position_collection_2[0, n])**2 +
            (position_collection_1[1, n]-position_collection_2[1, n])**2 +
            (position_collection_1[2, n]-position_collection_2[2, n])**2
        )**0.5
    return distance_collection

@njit(cache=True)
def sigma_to_shear(sigma):
    shear = np.zeros(sigma.shape)
    for n in range(shear.shape[1]):
        shear[0, n] = sigma[0, n]
        shear[1, n] = sigma[1, n]
        shear[2, n] = sigma[2, n] + 1
    return shear

@njit(cache=True)
def kappa_to_curvature(kappa, voronoi_dilatation):
    curvature = np.zeros(kappa.shape)
    for n in range(curvature.shape[1]):
        curvature[0, n] = kappa[0, n] / voronoi_dilatation[n]
        curvature[1, n] = kappa[1, n] / voronoi_dilatation[n]
        curvature[2, n] = kappa[2, n] / voronoi_dilatation[n]
    return curvature

class StaticRod:
    def __init__(
            self, rest_position, rest_director, rest_radius, shear_matrix, bend_matrix
        ):
        self.n_elements = rest_radius.shape[0]
        self.shear_matrix = shear_matrix.copy()
        self.bend_matrix = bend_matrix.copy()
        self.position_collection = rest_position.copy()
        self.director_collection = rest_director.copy()
        
        self.rest_radius = rest_radius.copy()
        self.rest_lengths = _batch_norm(
            _difference(self.position_collection)  # Position difference
        )
        self.rest_voronoi_lengths = 0.5 * (
            self.rest_lengths[1:] + self.rest_lengths[:-1]
        )

        self.lengths = np.zeros(self.n_elements)
        self.tangents = np.zeros((3, self.n_elements))
        self.radius = np.zeros(self.n_elements)

        self.dilatation = np.zeros(self.n_elements)
        self.voronoi_dilatation = np.zeros(self.n_elements-1)

        self.sigma = np.zeros((3, self.n_elements))
        self.kappa = np.zeros((3, self.n_elements-1))

        self._compute_geometry_from_state(
            self.position_collection, self.rest_lengths, self.rest_radius,
            self.lengths, self.tangents, self.radius
        )
        self._compute_all_dilatations(
            self.lengths,
            self.rest_lengths,
            self.rest_voronoi_lengths,
            self.dilatation,
            self.voronoi_dilatation,
        )
        self._compute_shear_stretch_strains(
            self.tangents,
            self.dilatation,
            self.director_collection,
            self.sigma,
        )
        self._compute_bending_twist_strains(
            self.director_collection, 
            self.rest_voronoi_lengths,
             self.kappa
        )

        self.rest_sigma = self.sigma.copy()
        self.rest_kappa = self.kappa.copy()

    @staticmethod
    @njit(cache=True)
    def _compute_bending_twist_strains(
        director_collection, rest_voronoi_lengths, kappa
    ):
        
        temp = _inv_rotate(director_collection)
        blocksize = rest_voronoi_lengths.shape[0]
        for k in range(blocksize):
            # kappa[0, k] = temp[0, k] / rest_voronoi_lengths[k]
            # kappa[1, k] = temp[1, k] / rest_voronoi_lengths[k]
            # kappa[2, k] = temp[2, k] / rest_voronoi_lengths[k]

            # The following should be the right implementation once the _inv_rotate is fixed
            kappa[0, k] = -temp[0, k] / rest_voronoi_lengths[k]
            kappa[1, k] = -temp[1, k] / rest_voronoi_lengths[k]
            kappa[2, k] = -temp[2, k] / rest_voronoi_lengths[k]

    @staticmethod
    @njit(cache=True)
    def _compute_shear_stretch_strains(
        tangents,
        dilatation,
        director_collection,
        sigma,
    ):
        # Quick trick : Instead of evaliation Q(et-d^3), use property that Q*d3 = (0,0,1), a constant
        z_vector = np.array([0.0, 0.0, 1.0]).reshape(3, -1)
        sigma[:] = dilatation * _batch_matvec(director_collection, tangents) - z_vector

    @staticmethod
    @njit(cache=True)
    def _compute_all_dilatations(
        lengths,
        rest_lengths,
        rest_voronoi_lengths,
        dilatation,
        voronoi_dilatation,
    ):
        for k in range(lengths.shape[0]):
            dilatation[k] = lengths[k] / rest_lengths[k]

        # Cmopute eq (3.4) from 2018 RSOS paper
        voronoi_lengths = _average(lengths)  # Position average

        # Cmopute eq (3.5) from 2018 RSOS paper
        for k in range(voronoi_lengths.shape[0]):
            voronoi_dilatation[k] = voronoi_lengths[k] / rest_voronoi_lengths[k]

    @staticmethod
    @njit(cache=True)
    def _compute_geometry_from_state(
        position_collection, rest_lengths, rest_radius, lengths, tangents, radius
    ):
        # Compute eq (3.3) from 2018 RSOS paper
        position_diff = _difference(position_collection)  # Position difference
        lengths[:] = _batch_norm(position_diff)
        for k in range(lengths.shape[0]):
            tangents[0, k] = position_diff[0, k] / lengths[k]
            tangents[1, k] = position_diff[1, k] / lengths[k]
            tangents[2, k] = position_diff[2, k] / lengths[k]
            # recalculate radius based on volume conservation
            radius[k] = rest_radius[k] * np.sqrt(rest_lengths[k]/lengths[k])

    def update_from_strain(self, sigma, kappa):
        self.sigma[:, :] = sigma.copy()
        self.kappa[:, :] = kappa.copy()
        self.static_pose_evolution(
            self.rest_lengths, self.sigma, self.kappa,
            self.position_collection, self.director_collection
        )
        self._compute_geometry_from_state(
            self.position_collection, self.rest_lengths, self.rest_radius,
            self.lengths, self.tangents, self.radius
        )
        self._compute_all_dilatations(
            self.lengths,
            self.rest_lengths,
            self.rest_voronoi_lengths,
            self.dilatation,
            self.voronoi_dilatation,
        )

    @staticmethod
    @njit(cache=True)
    def static_pose_evolution(
        rest_lengths, sigma, kappa,
        position_collection, director_collection
    ):
        shear = sigma_to_shear(sigma)
        for k in range(rest_lengths.shape[0]-1):
            next_position(
                director_collection[:, :, k],
                shear[:, k] * rest_lengths[k],
                position_collection[:, k:k+2]
                )
            next_director(
                kappa[:, k] * rest_lengths[k],
                director_collection[:, :, k:k+2]
                )

        next_position(
            director_collection[:, :, -1],
            shear[:, -1] * rest_lengths[-1],
            position_collection[:, -2:]
            )

    @classmethod
    def get_rod(cls, rest_cosserat_rod):
        return StaticRod(
            rest_cosserat_rod.position_collection,
            rest_cosserat_rod.director_collection,
            rest_cosserat_rod.radius,
            rest_cosserat_rod.shear_matrix,
            rest_cosserat_rod.bend_matrix
        )

@njit(cache=True)
def next_position(director, delta, positions):
    positions[:, 1] = positions[:, 0]
    for i in range(3):
        for j in range(3):
            positions[i, 1] += (
                director[j, i] * delta[j]
            )

@njit(cache=True)
def next_director(rotation, directors):
    # FIXME The following should be the right implementation once the _get_rotation_matrix is fixed
    # Rotation = _get_rotation_matrix(-1, rotation.reshape((3, 1)))[:, :, 0]
    
    Rotation = _get_rotation_matrix(1, rotation.reshape((3, 1)))[:, :, 0]
    for i in range(3):
        for j in range(3):
            directors[i, j, 1] = 0
            for k in range(3):
                directors[i, j, 1] += (
                    Rotation[i, k] * directors[k, j, 0]
                )



from coomm.actuations.muscles.muscle import *
from coomm.actuations.muscles.transverse_muscle import *
from coomm.actuations.muscles.longitudinal_muscle import *
from coomm.actuations.muscles.oblique_muscle import *



__doc__ = """
Oblique muscle implementation.
"""

from typing import Union

import numpy as np
from numba import njit

from coomm.actuations.muscles.muscle import MuscleForce


class ObliqueMuscle(MuscleForce):
    """ObliqueMuscle."""

    def __init__(
        self,
        muscle_init_angle: float,
        ratio_muscle_position: np.ndarray,
        rotation_number: float,
        rest_muscle_area: np.ndarray,
        max_muscle_stress: Union[float, np.ndarray],
        **kwargs
    ):
        """__init__.

        Parameters
        ----------
        muscle_init_angle : float
        ratio_muscle_position : np.ndarray
            shape: (3, n_element)
        rotation_number : float
        rest_muscle_area : np.ndarray
            shape: (n_element)
        max_muscle_stress : Union[float, np.ndarray]
            shape: (n_element)
        """
        n_elem = rest_muscle_area.shape[0]
        s = np.linspace(0, 1, n_elem + 1)
        s_muscle_position = (s[:-1] + s[1:]) / 2
        self.N = rotation_number  # TODO: Where is this used? If not outsideo of ObliqueMuscle, we shouldn't need self. Also we need better name than N.
        # fmt: off
        super().__init__(
            ratio_muscle_position=ratio_muscle_position * np.array([ \
                    np.cos(muscle_init_angle + 2 * np.pi * self.N * s_muscle_position),
                    np.sin(muscle_init_angle + 2 * np.pi * self.N * s_muscle_position),
                    np.zeros(n_elem)]),
            rest_muscle_area=rest_muscle_area,
            max_muscle_stress=max_muscle_stress,
            type_name="OM",
            **kwargs
        )
        # fmt: on



"""
Created on Dec. 17, 2021
@author: Heng-Sheng (Hanson) Chang
"""

from .povray_base import *
from .povray_frame import *
from .povray_camera import *
from .rod import *
from .muscles import *
from .sphere import *
from .cylinder import *

from .draw_sucker import *



"""
Created on Feb. 21, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from coomm._rod_tool import _material_to_lab

class POVRAYMuscle:
    def __init__(self, muscle_color, activation_color):
        self.muscle_color = np.array(muscle_color)
        self.activation_color = np.array(activation_color)

class POVRAYSlenderMuscle(POVRAYMuscle):
    def __init__(self, **kwargs):
        muscle_color = kwargs["muscle_color"]
        POVRAYMuscle.__init__(self, muscle_color, kwargs.get("activation_color", muscle_color))
        self.muscle_label = "// muscle data\n"

    def write_to(self, file, position_data, director_data, muscle_position_data, radius_data, muscle_activation=None, alpha=1.0):
        position = (position_data[:, :-1]+position_data[:, 1:])/2 
        position += _material_to_lab(director_data, muscle_position_data)
        position = self.adapted_position(position)
        radius = radius_data.copy()
        n_elements = position.shape[1]

        if muscle_activation is None:
            muscle_activation = np.ones(n_elements)

        start_index = 0
        end_index = n_elements - 2

        string = self.muscle_label
        string += "union{\n"

        for n in range(start_index, end_index-1):
            # string += "\tsphere{\n"
            # string += "\t\t<%f, %f, %f>, %f\n" % (
            #     position[0, n],
            #     position[1, n],
            #     position[2, n],
            #     radius[n]
            # )
            # string += "\t\ttexture{\n"
            # string += "\t\t\tpigment{ color rgb " + self.color_string 
            # string += " transmit %f }\n" % self.alpha_to_transmit(alpha*muscle_activation[n])
            # string += "\t\t\tfinish{ phong 1 }\n"
            # string += "\t\t}\n"
            # string += "\t\tscale<1, 1, 1>*%f\n" % self.scale
            # string += "\t}\n"

            string += "\tcone{\n"
            string += "\t\t<%f, %f, %f>, %f,\n" % (
                position[0, n], position[1, n], position[2, n], radius[n]
            )
            string += "\t\t<%f, %f, %f>, %f\n" % (
                position[0, n+1], position[1, n+1], position[2, n+1], radius[n+1]
            )
            string += "\t\ttexture{\n"
            string += "\t\t\tpigment{ color rgb " + self.color_string 
            string += " transmit %f }\n" % self.alpha_to_transmit(
                alpha*(muscle_activation[n]+muscle_activation[n+1])/2
            )
            string += "\t\t\tfinish{ phong 1 }\n"
            string += "\t\t}\n"
            string += "\t\tscale<1, 1, 1>*%f\n" % self.scale
            string += "\t}\n"

            # string += "\tsphere{\n"
            # string += "\t\t<%f, %f, %f>, %f\n" % (
            #     position[0, n+1],
            #     position[1, n+1],
            #     position[2, n+1],
            #     radius[n+1]
            # )
            # string += "\t\ttexture{\n"
            # string += "\t\t\tpigment{ color rgb " + self.color_string 
            # string += " transmit %f }\n" % self.alpha_to_transmit(alpha*muscle_activation[n+1])
            # string += "\t\t\tfinish{ phong 1 }\n"
            # string += "\t\t}\n"
            # string += "\t\tscale<1, 1, 1>*%f\n" % self.scale
            # string += "\t}\n"

        string += "}\n\n"

        file.writelines(string)

class POVRAYRingMuscle(POVRAYMuscle):
    def __init__(self, **kwargs):
        muscle_color = kwargs["muscle_color"]
        POVRAYMuscle.__init__(self, muscle_color, kwargs.get("activation_color", muscle_color))
        self.n_muscle_nodes = kwargs.get("n_muscle_nodes", 20)
        self.muscle_label = "// muscle data\n"
    
    def write_to(self, file, position_data, director_data, muscle_position_data, radius_data, muscle_activation=None, alpha=1.0):
        position = (position_data[:, :-1]+position_data[:, 1:])/2
        position_difference = position_data[:, :-1] - position_data[:, 1:]
        
        position += _material_to_lab(director_data, muscle_position_data)
        position = self.adapted_position(position)
        n_elements = position.shape[1]

        major_radius = radius_data.copy()
        minor_radius = np.linalg.norm(position_difference, axis=0)
        
        for n in range(n_elements):
            minor_radius[n] = np.min([major_radius[n]/2, minor_radius[n]])
        

        if muscle_activation is None:
            muscle_activation = np.ones(n_elements)

        string = self.muscle_label
        for n in range(n_elements-1):
            string += "sphere_sweep{\n"
            string += "\tb_spline %d" % (self.n_muscle_nodes+1)
            for nn in range(self.n_muscle_nodes+1):
                theta = nn/self.n_muscle_nodes*2*np.pi
                position_muscle = position[:, n] + (major_radius[n]-minor_radius[n]) * (
                    np.cos(theta)*director_data[0, :, n]+np.sin(theta)*director_data[1, :, n]
                )
                string += ",\n\t<%f, %f, %f>, %f" % (
                    position_muscle[0], position_muscle[1], position_muscle[2], minor_radius[n]
                )
            string += "\n\ttexture{\n"
            string += "\t\tpigment{ color rgb " + self.color_string 
            string += " transmit %f }\n" % self.alpha_to_transmit(alpha*muscle_activation[n])
            string += "\t\tfinish{ phong 1 }\n"
            string += "\t}\n"
            string += "\tscale<1, 1, 1>*%f\n" % self.scale
            string += "}\n"

        string += "\n"

        file.writelines(string)



"""
Created on Dec. 20, 2021
@author: Heng-Sheng (Hanson) Chang
"""

class POVRAYFrame:
    def __init__(self, included_files):
        self.included_files = included_files

    def write_included_files_to(self, file):
        for included_file in self.included_files:
            file.writelines('#include \"' + included_file + '\"\n')



__doc__ = """
Target Implementation
"""

import numpy as np

# from objects.object import Object

class Target:  # TODO: Maybe just use dataclass?
    def __init__(self, target_cost_weight):
        self.target_cost_weight = target_cost_weight



__doc__ = """
Longitudinal muscle model definition.
"""

from typing import Union

import numpy as np
from numba import njit

from coomm.actuations.muscles.muscle import MuscleForce


class LongitudinalMuscle(MuscleForce):
    """LongitudinalMuscle"""

    def __init__(
        self,
        muscle_init_angle: float,
        ratio_muscle_position: np.ndarray,
        rest_muscle_area: np.ndarray,
        max_muscle_stress: Union[float, np.ndarray],
        **kwargs
    ):
        """
        Initialize longitudinal muscle model.

        Parameters
        ----------
        muscle_init_angle : float
        ratio_muscle_position : np.ndarray
            shape: (3, n_element)
        rest_muscle_area : np.ndarray
            shape: (n_element)
        max_muscle_stress : Union[float, np.ndarray]
            shape: (n_element)
        """

        # FIXME: I'm not sure keeping ratio_muscle_position name after the operation is good idea.
        ratio_muscle_position = ratio_muscle_position * np.array(
            [[np.cos(muscle_init_angle)], [np.sin(muscle_init_angle)], [0]]
        )
        super().__init__(
            ratio_muscle_position=ratio_muscle_position,
            rest_muscle_area=rest_muscle_area,
            max_muscle_stress=max_muscle_stress,
            type_name="LM",
            **kwargs
        )



__doc__ = """
Director constraint object implementation
"""

import numpy as np

from coomm.objects.object import Object
from coomm.objects.target import Target

class DirectorConstraint(Object, Target):
    """DirectorConstraint.
    """

    def __init__(self, director, n_elements, cost_weight, target_cost_weight, **kwargs):
        """__init__.

        Parameters
        ----------
        director :
            director
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        target_cost_weight :
            target_cost_weight
        kwargs :
            kwargs
        """
        Object.__init__(self, n_elements, cost_weight)
        Target.__init__(self, target_cost_weight)
        self.director = director.copy()
        
        # self.director_flags = [True, True, True]
        # for i in range(len(self.director_flags)):
        #     if self.director[i, 0] is None:
        #         self.director_flags[i] = False

    def update_constraint(self, target_director, director):
        """update_constraint.

        Parameters
        ----------
        target_director :
            target_director
        director :
            director
        """
        self.update_director(target_director, director)

    # def update_position(self, position):
    #     self.position = position.copy()

    def update_director(self, target_director, director):
        """update_director.

        Parameters
        ----------
        target_director :
            target_director
        director :
            director
        """
        # for i, director_flag in enumerate(self.director_flags):
        #     self.director[i, :] = (
        #         target_director[i, :].copy() if director_flag else director[i, :].copy()
        #     ) 
        pass

    # @classmethod
    # def get_point_from_sphere(cls, sphere, n_elements, cost_weight):
    #     return Point(
    #         sphere.position_collection[:, 0].copy(),
    #         sphere.director_collection[:, :, 0].copy(),
    #         n_elements, cost_weight
    #     )

    def calculate_continuous_cost_gradient_wrt_position(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_position.

        Parameters
        ----------
        kwargs :
            kwargs
        """
        pass
    
    def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_director.

        Parameters
        ----------
        kwargs :
            kwargs
        """
        director = kwargs['director'].copy()
        vector = np.zeros((3, self.n_elements))
        skew_symmetric_matrix = (
            np.einsum('ijn,kjn->ikn', director, self.director) - 
            np.einsum('ijn,kjn->ikn', self.director, director)
        )
        vector[0, :] = skew_symmetric_matrix[1, 2, :]
        vector[1, :] = -skew_symmetric_matrix[0, 2, :]
        vector[2, :] = skew_symmetric_matrix[0, 1, :]
        self.cost_gradient.discrete.wrt_director[:, :] = (
            self.target_cost_weight['director'] * np.einsum('jik,jk->ik', director, vector)
        )
        # print("director")
        # print(director)
        # print("self.director")
        # print(self.director)
        # print("vector")
        # print(vector)
    
    def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_position.

        Parameters
        ----------
        kwargs :
            kwargs
        """
        pass
    
    def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_director.

        Parameters
        ----------
        kwargs :
            kwargs
        """
        pass

# TODO: What is this? Do we want to include or remove?
# class PointTarget(Point, Target):
#     def __init__(self, position, director, n_elements, cost_weight, target_cost_weight, **kwargs):
#         Point.__init__(self, position, director, n_elements, cost_weight)
#         Target.__init__(self, target_cost_weight)
#         self.director_cost_flag = kwargs.get('director_cost_flag', False)

#     @classmethod
#     def get_point_target_from_sphere(cls, sphere, n_elements, cost_weight, target_cost_weight, **kwargs):
#         return PointTarget(
#             sphere.position_collection[:, 0].copy(),
#             sphere.director_collection[:, :, 0].copy(),
#             n_elements, cost_weight, target_cost_weight, **kwargs
#         )
    
#     def calculate_continuous_cost_gradient_wrt_position(self, **kwargs):
#         pass
    
#     def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
#         pass
    
#     def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
#         position = 0.5*(kwargs['position'][:, -1]+kwargs['position'][:, -2])
#         self.cost_gradient.discrete.wrt_position[:, -1] = (
#             self.target_cost_weight['position'] * (position-self.position)
#         )
    
#     def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
#         director = kwargs['director'][:, :, -1]
#         vector = np.zeros(3)
#         skew_symmetric_matrix = director @ self.director.T - self.director @ director.T
#         vector[0] = skew_symmetric_matrix[1, 2]
#         vector[1] = -skew_symmetric_matrix[0, 2]
#         vector[2] = skew_symmetric_matrix[0, 1]
#         self.cost_gradient.discrete.wrt_director[:, -1] = (
#             self.target_cost_weight['director'] * director.T @ vector
#         )



__doc__ = """
Actuation tools
"""
__all__ = ['ApplyActuations', 'ContinuousActuation']


import numpy as np
from numba import njit

from elastica._linalg import _batch_cross
from elastica._calculus import quadrature_kernel, difference_kernel
from elastica.external_forces import inplace_addition
from elastica.external_forces import NoForces

from coomm._rod_tool import _lab_to_material, _material_to_lab, average2D

@njit(cache=True)
def _internal_to_external_load(
    director_collection, kappa, tangents,
    rest_lengths, rest_voronoi_lengths,
    dilatation, voronoi_dilatation,
    internal_force, internal_couple,
    external_force, external_couple
    ):

    external_force[:, :] = difference_kernel(
        _material_to_lab(director_collection, internal_force)
        )

    external_couple[:, :] = (
        difference_kernel(internal_couple) +
        quadrature_kernel(
            _batch_cross(kappa, internal_couple) * rest_voronoi_lengths
        ) +
        _batch_cross(
                _lab_to_material(director_collection, tangents * dilatation),
                internal_force
            )  * rest_lengths
    )

@njit(cache=True)
def _force_induced_couple(internal_forces, distance, internal_couples):
    internal_couples[:, :] = average2D(
        _batch_cross(distance, internal_forces)
    )

class ContinuousActuation:
    """ 
    Classes inherited from this base class should contain parameters and
    functions that describe and calculate forces / couples generated by
    actuators.
    """

    def __init__(self, n_elements: int, **kwargs):
        super().__init__()
        self.n_elements = n_elements
        self.internal_force = np.zeros((3, n_elements))       # material frame
        self.external_force = np.zeros((3, n_elements+1))     # lab frame
        self.internal_couple = np.zeros((3, n_elements-1))    # material frame
        self.external_couple = np.zeros((3, n_elements))      # material frame

    def reset_actuation(self,):
        """
        Set actuation force/torques to be zero.
        """
        self.internal_force[:, :] *= 0
        self.external_force[:, :] *= 0
        self.internal_couple[:, :] *= 0
        self.external_couple[:, :] *= 0

class ApplyActuations(NoForces):
    """ApplyActuations
    """

    def __init__(self, actuations, step_skip: int, callback_params_list: list):
        """
        TODO : need documentation on how to initialize

        Parameters
        ----------
        actuations :
        step_skip : int
        callback_params_list : Dictionary[list]
        """
        self.current_step = 0
        self.actuations = actuations
        self.every = step_skip
        self.callback_params_list = callback_params_list

    def apply_torques(self, system, time: np.float64 = 0.0):
        """apply_torques.

        Parameters
        ----------
        system :
        time : np.float64
        """
        for actuation in self.actuations:
            actuation(system)
            inplace_addition(
                system.external_forces, actuation.external_force
            )
            inplace_addition(
                system.external_torques, actuation.external_couple
            )

        self.make_callback()

    def make_callback(self):
        """make_callback.
        """
        if self.current_step % self.every == 0:
            self.callback_func(
                self.actuations,
                self.callback_params_list
            )
        self.current_step += 1

    def callback_func(self, actuations, callback_params_list):
        """callback_func.

        Parameters
        ----------
        actuations :
        callback_params_list :
        """
        for actuation, callback_params in zip(
            actuations, callback_params_list
        ):
            callback_params['internal_force'].append(
                actuation.internal_force.copy()
            )
            callback_params['internal_couple'].append(
                actuation.internal_couple.copy()
            )
            callback_params['external_force'].append(
                actuation.external_forces.copy()
            )
            callback_params['external_couple'].append(
                actuation.external_couple.copy()
            )



__doc__ = """
Frame base class
"""

import os
import matplotlib.pyplot as plt
from matplotlib import gridspec

from coomm._rendering_tool import check_folder

class FrameBase:
    """FrameBase.
    """

    def __init__(self, file_dict, fig_dict, gs_dict):
        """__init__.

        Parameters
        ----------
        file_dict :
        fig_dict :
        gs_dict :
        """
        self.figure_name = file_dict["figure_name"]
        self.folder_name = file_dict.get("folder_name", None)
        self.frame_count = 0
        self.fig_dict = fig_dict
        self.gs_dict = gs_dict
        self.fig = None
        self.gs = None

        if file_dict.get("check_folder_flag", True):
            check_folder(self.folder_name)
            file_dict['check_folder_flag'] = False

    def reset(self):
        """reset.
        """
        if self.fig is None:
            self.fig = plt.figure(**self.fig_dict)
            self.gs = gridspec.GridSpec(
                figure=self.fig,
                **self.gs_dict
            )

    def show(self,):
        """show.
        """
        plt.show()

    def save(self, show=False, frame_count=None):
        """save.

        Parameters
        ----------
        show :
        frame_count :
        """
        if self.folder_name is None:
            self.fig.savefig(self.figure_name)
        else:
            frame_count = (
                self.frame_count if frame_count is None else frame_count
            )
            self.fig.savefig(
                self.folder_name + "/" +
                self.figure_name.format(frame_count)
            )
            self.frame_count += 1
        if show:
            self.show()
        else:
            plt.close(self.fig)
        self.fig = None
        self.gs = None

    def movie(self, frame_rate, movie_name, start_number=0):
        """movie.

        Parameters
        ----------
        frame_rate :
        movie_name :
        start_number :
        """
        print("Creating movie:", movie_name+".mov")
        cmd = "ffmpeg -r {}".format(frame_rate)
        figure_name = self.figure_name.replace("{:", "%")
        figure_name = figure_name.replace("}", "")
        cmd += " -start_number {}".format(start_number)
        cmd += " -i " + self.folder_name + "/" + figure_name
        cmd += " -b:v 90M -c:v libx264 -pix_fmt yuv420p -f mov"
        cmd += " -y " + movie_name + ".mov"
        os.system(cmd)

# TODO: What is this??
# -vf "pad=ceil(iw/2)*2:ceil(ih/2)*2"



from coomm.actuations import *
from coomm.actuations.muscles import *
from coomm.algorithms import *
from coomm.forces import *
from coomm.frames import *
from coomm.objects import *
from coomm.povray import *
from coomm import *



__doc__ = """
Point object implementation
"""

import numpy as np

from coomm.objects.object import Object
from coomm.objects.target import Target

class Point(Object):
    """Point.
    """

    def __init__(self, position, director, n_elements, cost_weight):
        """__init__.

        Parameters
        ----------
        position :
            position
        director :
            director
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        """
        Object.__init__(self, n_elements, cost_weight)
        self.position = position.copy()
        self.director = director.copy()

    def update_pose_from_sphere(self, sphere):
        """update_pose_from_sphere.

        Parameters
        ----------
        sphere :
            sphere
        """
        self.update_position(sphere.position_collection[:, 0])
        self.update_director(sphere.director_collection[:, :, 0])

    def update_position(self, position):
        """update_position.

        Parameters
        ----------
        position :
            position
        """
        self.position = position.copy()

    def update_director(self, director):
        """update_director.

        Parameters
        ----------
        director :
            director
        """
        self.director = director.copy()

    @classmethod
    def get_point_from_sphere(cls, sphere, n_elements, cost_weight):
        """get_point_from_sphere.

        Parameters
        ----------
        sphere :
            sphere
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        """
        return Point(
            sphere.position_collection[:, 0].copy(),
            sphere.director_collection[:, :, 0].copy(),
            n_elements, cost_weight
        )


class PointTarget(Point, Target):
    """PointTarget.
    """

    def __init__(self, position, director, n_elements, cost_weight, target_cost_weight, **kwargs):
        """__init__.

        Parameters
        ----------
        position :
            position
        director :
            director
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        target_cost_weight :
            target_cost_weight
        kwargs :
            kwargs
        """
        Point.__init__(self, position, director, n_elements, cost_weight)
        Target.__init__(self, target_cost_weight)
        self.director_cost_flag = kwargs.get('director_cost_flag', False)

    @classmethod
    def get_point_target_from_sphere(cls, sphere, n_elements, cost_weight, target_cost_weight, **kwargs):
        """get_point_target_from_sphere.

        Parameters
        ----------
        sphere :
            sphere
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        target_cost_weight :
            target_cost_weight
        kwargs :
            kwargs
        """
        return PointTarget(
            sphere.position_collection[:, 0].copy(),
            sphere.director_collection[:, :, 0].copy(),
            n_elements, cost_weight, target_cost_weight, **kwargs
        )
    
    def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_position.

        Parameters
        ----------
        kwargs :
            kwargs
        """
        position = 0.5*(kwargs['position'][:, -1]+kwargs['position'][:, -2])
        self.cost_gradient.discrete.wrt_position[:, -1] = (
            self.target_cost_weight['position'] * (position-self.position)
        )
    
    def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_director.

        Parameters
        ----------
        kwargs :
            kwargs
        """
        director = kwargs['director'][:, :, -1]
        vector = np.zeros(3)
        skew_symmetric_matrix = director @ self.director.T - self.director @ director.T
        vector[0] = skew_symmetric_matrix[1, 2]
        vector[1] = -skew_symmetric_matrix[0, 2]
        vector[2] = skew_symmetric_matrix[0, 1]
        self.cost_gradient.discrete.wrt_director[:, -1] = (
            self.target_cost_weight['director'] * director.T @ vector
        )



"""
Created on Dec. 17, 2021
@author: Heng-Sheng (Hanson) Chang
"""

from numba import njit
import numpy as np

from coomm._rendering_tool import (
    process_position, process_director
)

class POVRAYBase:
    def __init__(self, **kwargs):
        self.color_string = self.to_color_string(kwargs.get("color", [0.45, 0.39, 1.0]))
        self.scale = kwargs.get("scale", 16)
        self.rotation_matrix = kwargs.get("rotation_matrix", np.eye(3))
        self.offset = kwargs.get("offset", np.zeros(3))
    
    @staticmethod
    def alpha_to_transmit(alpha):
        return 1-alpha

    def to_color_string(self, color):
        string = color if isinstance(color, str) else (
            "<%f, %f, %f>" % (color[0], color[1], color[2])
        )
        return string

    def adapted_position(self, position):
        return process_position(position, self.offset, self.rotation_matrix)

    def adapted_director(self, director):
        return process_director(director, self.rotation_matrix)
        



"""
Created on Feb. 21, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
from coomm.povray.povray_base import POVRAYBase

class POVRAYRod(POVRAYBase):
    def __init__(self, **kwargs):
        POVRAYBase.__init__(self, **kwargs)

    def write_to(self, file, position_data, radius_data, alpha):
        position = self.adapted_position(
            (position_data[:, :-1]+position_data[:, 1:])/2 
        )
        radius = radius_data
        n_elements = radius.shape[0]

        string = "// rod data\n"
        string += "sphere_sweep{\n\tb_spline %d" % n_elements
        for n in range(n_elements):
            string += (",\n\t<%f, %f, %f>, %f" % (position[0, n], position[1, n], position[2, n], radius[n]))
        string += "\n\ttexture{\n"
        string += "\t\tpigment{ color rgb" + self.color_string + " transmit %f }\n" % self.alpha_to_transmit(alpha)
        string += "\t\tfinish{ phong 1 }\n\t}\n"
        string += "\tscale<1, 1, 1>*%f\n}\n\n" % self.scale

        file.writelines(string)



__doc__ = """
Rigid-body frame implementation
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from coomm.frames.frame import FrameBase
from coomm.frames.frame_tools import (
    base_colors,
    default_label_fontsize,
)

from coomm._rendering_tool import (
    process_position, process_director
)

# TODO: Combine into class maybe
rigidbody_color = base_colors['y']

class RigidbodyFrame(FrameBase):
    """RigidbodyFrame.
    """

    def __init__(self, file_dict, fig_dict, gs_dict, **kwargs):
        """__init__.

        Parameters
        ----------
        file_dict :
        fig_dict :
        gs_dict :
        """
        FrameBase.__init__(
            self,
            file_dict=file_dict,
            fig_dict=fig_dict,
            gs_dict=gs_dict
        )
        
        self.ax_main_info = kwargs["ax_main_info"]
        self.ax_main_indices = self.ax_main_info["indices"]
        self.ax_main_3d_flag = not self.ax_main_info.get("planner_flag", True)
        self.plot_rod = (
            self.plot_rod3d if self.ax_main_3d_flag else self.plot_rod2d
        )

        self.fontsize = kwargs.get("fontsize", default_label_fontsize)
        self.rigidbody_color = kwargs.get("rigidbody_color", rigidbody_color)
        self.offset = kwargs.get("offset", np.zeros(3))
        self.rotation = kwargs.get("rotation", np.identity(3))
        self.reference_total_length = 1
        self.reference_configuration_flag = False
        RigidbodyFrame.set_n_elems(self, kwargs.get("n_elems", 1))

    def set_n_elems(self, n_elems):
        """set_n_elems.

        Parameters
        ----------
        n_elems :
        """
        self.n_elems = n_elems

    def reset(self,):
        """reset.
        """
        FrameBase.reset(self,)
        
        if self.ax_main_3d_flag:
            self.ax_main = self.fig.add_subplot(
                self.gs[
                    self.ax_main_indices[0],
                    self.ax_main_indices[1]
                ],
                projection='3d'
            )
        else:
            self.ax_main = self.fig.add_subplot(
                self.gs[
                    self.ax_main_indices[0],
                    self.ax_main_indices[1]
                ]
            )

    def plot_rigidybody2d(self, position, director, radius, color=None):
        """plot_rigidybody2d.

        Parameters
        ----------
        position :
        director :
        radius :
        color :
        """
        return self.ax_main

    def plot_rigidybody3d(self, position, director, radius, color=None):
        """plot_rigidybody3d.

        Parameters
        ----------
        position :
        director :
        radius :
        color :
        """
        return self.ax_main

    def set_ax_main_lim(
        self, 
        x_lim=[-1.1, 1.1], # FIXME: Never use mutable object for default
        y_lim=[-1.1, 1.1],
        z_lim=[-1.1, 1.1]
    ):
        """set_ax_main_lim.

        Parameters
        ----------
        x_lim :
        y_lim :
        z_lim :
        """
        self.ax_main.set_xlim(x_lim)
        self.ax_main.set_ylim(y_lim)
        if self.ax_main_3d_flag:
            self.ax_main.set_zlim(z_lim)

    def set_labels(self, time=None):
        """set_labels.

        Parameters
        ----------
        time :
        """
        if time is not None:
            self.ax_main.set_title(
                "time={:.2f} [sec]".format(time), 
                fontsize=self.fontsize
            )
        
        self.ax_main.legend()



__doc__ = """
Cylinder object implementation
"""

import numpy as np

from coomm.objects.object import Object
from coomm.objects.target import Target

class Cylinder(Object):
    """Cylinder.
    """

    def __init__(self, position, director, radius, length, n_elements, cost_weight):
        """__init__.

        Parameters
        ----------
        position :
            position
        director :
            director
        radius :
            radius
        length :
            length
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        """
        Object.__init__(self, n_elements, cost_weight)
        self.position = position.copy()
        self.director = director.copy()
        self.radius = radius
        self.length = length

    def update_pose_from_sphere(self, sphere):
        """update_pose_from_sphere.

        Parameters
        ----------
        sphere :
            sphere
        """
        self.update_position(sphere.position_collection[:, 0])
        self.update_director(sphere.director_collection[:, :, 0])

    def update_position(self, position):
        """update_position.

        Parameters
        ----------
        position :
            position
        """
        self.position = position.copy()

    def update_director(self, director):
        """update_director.

        Parameters
        ----------
        director :
            director
        """
        self.director = director.copy()

    @classmethod
    def get_cylinder(cls, cylinder, n_elements, cost_weight):
        """get_cylinder.

        Parameters
        ----------
        cylinder :
            cylinder
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        """
        return cls(
            cylinder.position_collection[:, 0].copy(),
            cylinder.director_collection[:, :, 0].copy(),
            n_elements, cost_weight
        )

    def calculate_continuous_cost_gradient_wrt_position(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_position.
        """
        position = 0.5*(kwargs['position'][:, :-1]+kwargs['position'][:, 1:])
        radius = kwargs['radius']
        position_diff = position-self.position[:, None]
        vertical_dist = np.einsum("ij, i -> j", position_diff, self.director[2, :])
        horizontal_position_diff = position_diff - vertical_dist * self.director[2, :, None]
        position_dist = np.linalg.norm(horizontal_position_diff, axis=0)
        adjust_distance_ratio = (position_dist-(radius+self.radius))/position_dist
        adjust_distance_ratio[adjust_distance_ratio>0] = 0
        adjust_distance_ratio[vertical_dist>self.length/2] = 0
        adjust_distance_ratio[-vertical_dist>self.length/2] = 0
        self.cost_gradient.continuous.wrt_position[:, :] = (
            self.cost_weight['position'] * horizontal_position_diff * adjust_distance_ratio
        )

    def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_director.
        """
        pass
    
    def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_position.
        """
        pass
    
    def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_director.
        """
        pass

class CylinderTarget(Cylinder, Target):
    """CylinderTarget.
    """

    def __init__(self, position, director, radius, length, n_elements, cost_weight, target_cost_weight, **kwargs):
        """__init__.

        Parameters
        ----------
        position :
        director :
        radius :
        length :
        n_elements :
        cost_weight :
        target_cost_weight :
        """
        Cylinder.__init__(self, position, director, radius, length, n_elements, cost_weight)
        Target.__init__(self, target_cost_weight)
        self.director_cost_flag = kwargs.get('director_cost_flag', False)

    @classmethod
    def get_cylinder(cls, cylinder, n_elements, cost_weight, target_cost_weight, **kwargs):
        """get_cylinder.

        Parameters
        ----------
        cylinder :
            cylinder
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        target_cost_weight :
            target_cost_weight
        """
        return CylinderTarget(
            cylinder.position_collection[:, 0].copy(),
            cylinder.director_collection[:, :, 0].copy(),
            cylinder.radius,
            cylinder.length,
            n_elements, cost_weight, target_cost_weight, **kwargs
        )
    
    def calculate_continuous_cost_gradient_wrt_position(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_position.
        """
        Cylinder.calculate_continuous_cost_gradient_wrt_position(self, **kwargs)
        position = 0.5*(kwargs['position'][:, :-1]+kwargs['position'][:, 1:])
        radius = kwargs['radius']
        position_diff = position-self.position[:, None]
        
        # vertical_dist = np.einsum("ij, i -> j", position_diff, self.director[2, :])
        # horizontal_position_diff = position_diff - vertical_dist * self.director[2, :, None]
        # position_dist = np.linalg.norm(horizontal_position_diff, axis=0)
        # adjust_distance_ratio = (position_dist-(radius+self.radius))/position_dist
        # adjust_distance_ratio[adjust_distance_ratio<0] = 0
        # self.cost_gradient.continuous.wrt_position[:, :] += (
        #     self.cost_weight['position'] * horizontal_position_diff * adjust_distance_ratio
        # )

        position_dist = np.linalg.norm(position_diff, axis=0)
        adjust_distance_ratio = (position_dist-(radius+self.radius))/position_dist
        
        adjust_distance_ratio[adjust_distance_ratio<0] = 0
        self.cost_gradient.continuous.wrt_position[:, :] += (
            self.target_cost_weight['position'] * position_diff * adjust_distance_ratio
        )
    
    """ The first method """
    # def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
    #     """calculate_continuous_cost_gradient_wrt_director.
    #     """
    #     position = 0.5*(kwargs['position'][:, :-1]+kwargs['position'][:, 1:])
    #     director = kwargs['director'][:, :, :]
    #     n_elems = director.shape[2]
    #     vector = np.zeros((3, n_elems))
    #     coefficient = np.zeros(n_elems)
    #     # for n in range(n_elems):
    #     #     skew_symmetric_matrix = director[:, :, n] @ self.director.T - self.director @ director[:, :, n].T
    #     #     vector[0, n] = skew_symmetric_matrix[1, 2]
    #     #     vector[1, n] = -skew_symmetric_matrix[0, 2]
    #     #     vector[2, n] = skew_symmetric_matrix[0, 1]
    #     #     self.cost_gradient.continuous.wrt_director[:, n] = (
    #     #         self.target_cost_weight['director'][n] * director[:, :, n].T @ vector[:, n]
    #     #     )
        
    #     for n in range(n_elems):
    #         vector[1, n] = - np.dot(director[2, :, n], self.director[2, :])
    #         vector[2, n] = np.dot(director[1, :, n], self.director[2, :])
    #         coefficient[n] = np.dot(director[0, :, n], self.director[2, :])
    #         self.cost_gradient.continuous.wrt_director[:, n] = (
    #             self.target_cost_weight['director'][n] * coefficient[n] * director[:, :, n].T @ vector[:, n]
    #         )

    #     vector = np.zeros((3, n_elems))
    #     coefficient = np.zeros(n_elems)
    #     for n in range(n_elems):
    #         position_diff = self.position - position[:, n]
    #         position_diff = position_diff / np.linalg.norm(position_diff)
    #         vector[1, n] = - np.dot(director[2, :, n], position_diff)
    #         vector[2, n] = np.dot(director[1, :, n], position_diff)
    #         coefficient[n] = min(np.dot(self.director[0, :], position_diff), 0)
    #         self.cost_gradient.continuous.wrt_director[:, n] += (
    #             self.target_cost_weight['director'][n] * coefficient[n] * director[:, :, n].T @ vector[:, n]
    #         )
    def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_director.
        """
        position = 0.5*(kwargs['position'][:, :-1]+kwargs['position'][:, 1:])
        director = kwargs['director'][:, :, :]
        n_elems = director.shape[2]

        # vector = np.zeros((3, n_elems))
        # coefficient = np.zeros(n_elems)
        # for n in range(n_elems):
        #     skew_symmetric_matrix = director[:, :, n] @ self.director.T - self.director @ director[:, :, n].T
        #     vector[0, n] = skew_symmetric_matrix[1, 2]
        #     vector[1, n] = -skew_symmetric_matrix[0, 2]
        #     vector[2, n] = skew_symmetric_matrix[0, 1]
        #     self.cost_gradient.continuous.wrt_director[:, n] = (
        #         self.target_cost_weight['director'][n] * director[:, :, n].T @ vector[:, n]
        #     )
        
        # for n in range(n_elems):
        #     vector[1, n] = - np.dot(director[2, :, n], self.director[2, :])
        #     vector[2, n] = np.dot(director[1, :, n], self.director[2, :])
        #     coefficient[n] = np.dot(director[0, :, n], self.director[2, :])
        #     self.cost_gradient.continuous.wrt_director[:, n] = (
        #         self.target_cost_weight['director'][n] * coefficient[n] * director[:, :, n].T @ vector[:, n]
        #     )



        """ The second method """
        vector = np.zeros((3, n_elems))
        coefficient = np.zeros(n_elems)
        for n in range(n_elems):
            position_diff = self.position - position[:, n]
            position_diff = position_diff / np.linalg.norm(position_diff)
            vector[1, n] = - np.dot(director[2, :, n], -position_diff)
            vector[2, n] = np.dot(director[1, :, n], -position_diff)
            # coefficient[n] = 1-np.dot(director[0, :, n], position_diff)
            coefficient[n] = 1
            self.cost_gradient.continuous.wrt_director[:, n] = (
                self.target_cost_weight['director'][n] * coefficient[n] * director[:, :, n].T @ vector[:, n]
            )

        """ The third method """
        # vector = np.zeros((3, n_elems))
        # coefficient = np.zeros(n_elems)
        # for n in range(n_elems):
        #     position_diff = self.position - position[:, n]
        #     position_diff = position_diff - np.dot(position_diff, self.director[2, :]) * self.director[2, :]
        #     position_diff = position_diff / np.linalg.norm(position_diff)
        #     vector[1, n] = - np.dot(director[2, :, n], -position_diff)
        #     vector[2, n] = np.dot(director[1, :, n], -position_diff)
        #     # coefficient[n] = 1 - np.dot(director[0, :, n], position_diff)
        #     coefficient[n] = 1 
        #     # print(position_diff)
        #     self.cost_gradient.continuous.wrt_director[:, n] = (
        #         self.target_cost_weight['director'][n] * coefficient[n] * director[:, :, n].T @ vector[:, n]
        #     )

    def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_position.
        """
        # position = 0.5*(kwargs['position'][:, -1]+kwargs['position'][:, -2])
        # self.cost_gradient.discrete.wrt_position[:, -1] = (
        #     self.target_cost_weight['position'] * (position-self.position)
        # )
        pass
    
    def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_director.
        """
        # director = kwargs['director'][:, :, -1]
        # vector = np.zeros(3)
        # skew_symmetric_matrix = director @ self.director.T - self.director @ director.T
        # vector[0] = skew_symmetric_matrix[1, 2]
        # vector[1] = -skew_symmetric_matrix[0, 2]
        # vector[2] = skew_symmetric_matrix[0, 1]
        # self.cost_gradient.discrete.wrt_director[:, -1] = (
        #     self.target_cost_weight['director'] * director.T @ vector
        # )
        pass



__doc__ = """
Base object module
"""

import numpy as np

class Object:  # FIXME: To general name.
    """Object.
    """

    def __init__(self, n_elements, cost_weight=None):
        """__init__.

        Parameters
        ----------
        n_elements :
            n_elements
        cost_weight :
            cost_weight
        """
        self.n_elements = n_elements
        self.cost = Cost(n_elements)
        self.cost_gradient = CostGradient(n_elements)
        self.cost_weight = cost_weight

    def __call__(self, **kwargs): # FIXME: Maybe avoid using __call__?
        """__call__
        """
        # All cost related terms should be calculated once this function is called
        # including running cost, terminal cost (might not just on terminal it should 
        # be extend to any or some specific s as well), running cost gradient and 
        # terminal cost gradient (similarly, should be extended to any or some specific s)
        self.calculate_cost_gradient(**kwargs)

    # def calculate_cost(self, **kwargs):
    #     # should return all cost terms including running cost and terminal cost
    #     return NotImplementedError

    def calculate_cost_gradient(self, **kwargs):
        """calculate_cost_gradient.
        """
        self.calculate_continuous_cost_gradient_wrt_position(**kwargs)
        self.calculate_continuous_cost_gradient_wrt_director(**kwargs)
        self.calculate_discrete_cost_gradient_wrt_position(**kwargs)
        self.calculate_discrete_cost_gradient_wrt_director(**kwargs)

    def calculate_continuous_cost_gradient_wrt_position(self, **kwargs):
        return NotImplementedError
    
    def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
        return NotImplementedError
    
    def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
        return NotImplementedError
    
    def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
        return NotImplementedError

class Cost:
    """Cost.
    """

    def __init__(self, n_elements):
        """__init__.

        Parameters
        ----------
        n_elements :
            n_elements
        """
        self.continuous = WRT_Pose(n_elements, n_elements, dim=1)
        self.discrete = WRT_Pose(n_elements, n_elements, dim=1)

    def reset(self,):
        """reset.
        """
        self.continuous.reset()
        self.discrete.reset()

class CostGradient():
    """CostGradient.
    """

    def __init__(self, n_elements):
        """__init__.

        Parameters
        ----------
        n_elements :
            n_elements
        """
        self.continuous = WRT_Pose(n_elements, n_elements)
        self.discrete = WRT_Pose(n_elements, n_elements)

    def reset(self,):
        """reset.
        """
        self.continuous.reset()
        self.discrete.reset()

    def add(self, other):
        """add.

        Parameters
        ----------
        other :
            other
        """
        self.continuous.add(other.continuous)
        self.discrete.add(other.discrete)

class WRT_Pose():
    """WRT_Pose.
    """

    def __init__(self, n_elements_for_position, n_elements_for_director, dim=3):
        """__init__.

        Parameters
        ----------
        n_elements_for_position :
            n_elements_for_position
        n_elements_for_director :
            n_elements_for_director
        dim :
            dim
        """
        if dim ==1:
            self.wrt_position = np.zeros(n_elements_for_position)
            self.wrt_director = np.zeros(n_elements_for_director)
        else:
            self.wrt_position = np.zeros((dim, n_elements_for_position))
            self.wrt_director = np.zeros((dim, n_elements_for_director))
    
    def reset(self,):
        """reset.
        """
        self.wrt_position *= 0
        self.wrt_director *= 0
    
    def add(self, other):
        """add.

        Parameters
        ----------
        other :
            other
        """
        self.wrt_position += other.wrt_position
        self.wrt_director += other.wrt_director

class Objects(Object): # FIXME: we should be clear on naming.
    """Objects.
    """

    def __init__(self, objects):
        """__init__.

        Parameters
        ----------
        objects :
            objects
        """
        Object.__init__(self, objects[0].n_elements)
        self.objects = objects
    
    def __getitem__(self, key):
        if isinstance(key, slice):
            return Objects(
                [self.objects[k] 
                for k in range(*key.indices(len(self.objects)))
                ]
            )
        return self.objects[key]
    
    def __setitem__(self, key, value):
        if isinstance(key, slice):
            for k, v in zip(
                range(*key.indices(len(self.objects))), value
            ):
                self.objects[k] = v
        else:
            self.objects[key] = value
    
    def append(self, value):
        """append.

        Parameters
        ----------
        value :
            value
        """
        self.objects.append(value)

    def __call__(self, **kwargs):
        self.reset()
        for obj in self.objects:
            obj(**kwargs)
            self.cost_gradient.add(obj.cost_gradient)

    def calculate_cost_gradient(self, **kwargs):
        """calculate_cost_gradient.
        """
        pass

    def calculate_continuous_cost_gradient_wrt_position(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_position.
        """
        pass
    
    def calculate_continuous_cost_gradient_wrt_director(self, **kwargs):
        """calculate_continuous_cost_gradient_wrt_director.
        """
        pass
    
    def calculate_discrete_cost_gradient_wrt_position(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_position.
        """
        pass
    
    def calculate_discrete_cost_gradient_wrt_director(self, **kwargs):
        """calculate_discrete_cost_gradient_wrt_director.
        """
        pass



"""
Created on Dec. 20, 2021
@author: Heng-Sheng (Hanson) Chang
"""

class POVRAYCamera:
    def __init__(self, position, look_at, angle, rotate=[0.0, 0.0, 0.0], translate=[0.0, 0.0, 0.0], sky=[0.0, 0.0, 1.0], sun_position=[-1500, -1000, 2000], **kwargs):
        self.camera_position = "<%f, %f, %f>" % (position[0], position[1], position[2])
        self.camera_look_at = "<%f, %f, %f>" % (look_at[0], look_at[1], look_at[2])
        self.camera_angle = "%f" % (angle)
        self.camera_sky = "<%f, %f, %f>" % (sky[0], sky[1], sky[2])
        self.camera_rotate = "<%f, %f, %f>" % (rotate[0], rotate[1], rotate[2])
        self.camera_translate = "<%f, %f, %f>" % (translate[0], translate[1], translate[2])
        self.sun_position = "<%f, %f, %f>" % (sun_position[0], sun_position[1], sun_position[2])
        
        self.starting_string = '//--------------------------------------------------\n'
        self.ending_sting ='\n'

        self.default_setting = self.starting_string
        self.default_setting += '#version 3.6; // 3.7;\n'
        self.default_setting += 'global_settings{ assumed_gamma 1.0 }\n'
        self.default_setting += '#default{ finish{ ambient 0.1 diffuse 0.9 }}\n'
        self.default_setting += self.ending_sting

        self.default_included_lib_files = self.starting_string
        self.default_included_lib_files += '#include \"colors.inc\"\n'
        self.default_included_lib_files += '#include \"textures.inc\"\n'
        self.default_included_lib_files += '#include \"glass.inc\"\n'
        self.default_included_lib_files += '#include \"metals.inc\"\n'
        self.default_included_lib_files += '#include \"golds.inc\"\n'
        self.default_included_lib_files += '#include \"stones.inc\"\n'
        self.default_included_lib_files += '#include \"woods.inc\"\n'
        self.default_included_lib_files += '#include \"shapes.inc\"\n'
        self.default_included_lib_files += '#include \"shapes2.inc\"\n'
        self.default_included_lib_files += '#include \"functions.inc\"\n'
        self.default_included_lib_files += '#include \"math.inc\"\n'
        self.default_included_lib_files += '#include \"transforms.inc\"\n'
        self.default_included_lib_files += self.ending_sting

        self.camera_setting = self.starting_string
        self.camera_setting += '#declare Camera_Position  = ' + self.camera_position + ';\n'
        self.camera_setting += '#declare Camera_Look_At   = ' + self.camera_look_at + ';\n'
        self.camera_setting += '#declare Camera_Sky       = ' + self.camera_sky + ';\n'
        self.camera_setting += '#declare Camera_Angle     = ' + self.camera_angle + ';\n'
        self.camera_setting += '#declare Camera_Rotate    = ' + self.camera_rotate + ';\n'
        self.camera_setting += '#declare Camera_Translate = ' + self.camera_translate + ';\n'
        self.camera_setting += 'camera{ location  Camera_Position\n'
        self.camera_setting += '        right     -x*image_width/image_height\n'
        self.camera_setting += '        sky       Camera_Sky\n'
        self.camera_setting += '        angle     Camera_Angle\n'
        self.camera_setting += '        look_at   Camera_Look_At\n'
        self.camera_setting += '        rotate    Camera_Rotate\n'
        self.camera_setting += '        translate Camera_Translate\n'
        self.camera_setting += '}\n\n'
        self.camera_setting += '#declare Sun_Position = ' + self.sun_position + ';\n'
        self.camera_setting += 'light_source{ Sun_Position color White}                      // sun light\n'
        self.camera_setting += 'light_source{ Camera_Position color rgb<0.9, 0.9, 1.0>*0.1}  // flash light\n'
        self.camera_setting += 'background { color White }\n'
        self.camera_setting += self.ending_sting

        if kwargs.get('floor', False):
            self.camera_setting += 'declare RasterScale = 2.0;\n'
            self.camera_setting += 'declare RasterHalfLine  = 0.05;\n'
            self.camera_setting += 'declare RasterHalfLineZ = 0.05;\n'
            self.camera_setting += 'declare Raster_Color = <1,1,1>*0.2;\n'
            self.camera_setting += '#macro Raster(RScale, HLine)\n'
            self.camera_setting += 'pigment{\n'
            self.camera_setting += '\tgradient x scale RScale\n'
            self.camera_setting += '\tcolor_map{\n'
            self.camera_setting += '\t\t[0.000   color rgb Raster_Color]\n'
            self.camera_setting += '\t\t[0+HLine color rgb Raster_Color]\n'
            self.camera_setting += '\t\t[0+HLine color rgbt<1,1,1,1>]\n'
            self.camera_setting += '\t\t[1-HLine color rgbt<1,1,1,1>]\n'
            self.camera_setting += '\t\t[1-HLine color rgb Raster_Color]\n'
            self.camera_setting += '\t\t[1.000   color rgb Raster_Color]\n'
            self.camera_setting += '\t}// end of color_map\n'
            self.camera_setting += '}// end of pigment\n'
            self.camera_setting += '#end// of Raster(RScale, HLine)-macro\n'
            self.camera_setting += 'plane { <0,0,-1>, 1.0 \n'
            self.camera_setting += '\ttexture { pigment{color White}\n'
            self.camera_setting += '\t\tfinish {ambient 0.45 diffuse 0.85}}\n'
            self.camera_setting += '\ttexture { Raster(RasterScale,RasterHalfLine ) rotate<0,0,0> }\n'
            self.camera_setting += '\ttexture { Raster(RasterScale,RasterHalfLineZ) rotate<0,0,90>}\n'
            self.camera_setting += '\trotate<0,0,30>\n'
            self.camera_setting += '}\n'
    
    def write_to(self, file,):
        file.writelines(self.default_setting)
        file.writelines(self.default_included_lib_files)
        file.writelines(self.camera_setting)


__doc__ = """
Muscle frame imple
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
from mpl_toolkits.mplot3d import Axes3D

from coomm.frames.frame import FrameBase
from coomm.frames.frame_tools import (
    default_colors,
    default_label_fontsize,
    change_box_to_arrow_axes,
    change_box_to_only_x_line_ax,
    change_box_to_only_y_arrow_ax,
    add_y_ticks
)

from coomm._rendering_tool import (
    process_position, process_director
)

# TODO: plot muscles in the ax_main, which needs 
# process_position and process_director from frame_tools

# TODO: Maybe combine default color/parameters into class?
muscle_length_color = default_colors['tab:brown']
muscle_force_weight_color = default_colors['tab:blue']

muscles_color = dict(
    TM=default_colors['tab:green'],
    LM=default_colors['tab:red'],
    OM=default_colors['tab:cyan'],
)


class MuscleFrameBase(FrameBase):
    """MuscleFrameBase.
    """

    def __init__(self, file_dict, fig_dict, gs_dict, **kwargs):
        """__init__.

        Parameters
        ----------
        file_dict :
        fig_dict :
        gs_dict :
        """
        FrameBase.__init__( # Try to use super
            self,
            file_dict=file_dict,
            fig_dict=fig_dict,
            gs_dict=gs_dict
        )

        self.fontsize = kwargs.get("fontsize", default_label_fontsize)
        self.offset = kwargs.get("offset", np.zeros(3))
        self.rotation = kwargs.get("rotation", np.identity(3))
        MuscleFrameBase.set_ref_setting( # TODO:.. This part look weird
            self,
            reference_total_length=kwargs.get("reference_total_length", 1),
            n_elems=kwargs.get("n_elems", 100)
        )

        self.ax_muscles_info = kwargs["ax_muscles_info"]

    def set_n_elems(self, n_elems):
        """set_n_elems.

        Parameters
        ----------
        n_elems :
        """
        self.n_elems = n_elems
        self.s = np.linspace(0, 1, self.n_elems+1)

    def set_ref_setting(self, reference_total_length, n_elems):
        """set_ref_setting.

        Parameters
        ----------
        reference_total_length :
        n_elems :
        """
        self.reference_total_length = reference_total_length
        MuscleFrameBase.set_n_elems(self, n_elems)

    def ax_muscles_reset(self):
        """ax_muscles_reset.
        """
        ax = self.fig.add_subplot(
            self.gs[
                self.ax_muscles_info['indices'][0],
                self.ax_muscles_info['indices'][1]
            ]
        )
        for spine in ax.spines.values():
            spine.set_visible(False)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        self.ax_muscles = ax
        return self.ax_muscles

    def reset(self, axes_muscle_info=None):
        """reset.

        Parameters
        ----------
        axes_muscle_info :
        """
        FrameBase.reset(self,)
        if axes_muscle_info is None:
            return self.ax_muscles_reset()

        axes_muscle = []
        axes_muscle_activation = []
        axes_muscle_length = []
        axes_muscle_force_weight = []

        for i, axes_muscle_indices_i in enumerate(axes_muscle_info['indices']):
            muscle_group_yticklabels_showflag = axes_muscle_info['yticklabels_showflag'][i]
            ax = self.fig.add_subplot(
                self.gs[
                    axes_muscle_indices_i['activation'][0],
                    axes_muscle_indices_i['activation'][1]
                ],
                xlim=[-0.05, 1.15],
                ylim=[-0.1, 1.1],
            )
            ax = change_box_to_arrow_axes(
                self.fig, ax,
                xaxis_ypos=-0.05,
                x_offset=[0.05, -0.05]
            )
            ax.tick_params(axis='x', direction='in', length=4, width=1, pad=0)
            ax.tick_params(axis='y', direction='in', length=4, width=1, pad=1)
            if muscle_group_yticklabels_showflag['activation'] is not True:
                ax.set_yticklabels([])
            axes_muscle_activation.append(ax)

            axes_muscle_length.append([])
            axes_muscle_force_weight.append([])
            for j, axes_muscle_indices_i_j in enumerate(axes_muscle_indices_i['others']):
                ax = self.fig.add_subplot(
                    self.gs[
                        axes_muscle_indices_i_j[0],
                        axes_muscle_indices_i_j[1]
                    ],
                    xlim=[-0.05, 1.15],
                    ylim=[-0.2, 2.2],
                )
                ax = change_box_to_only_x_line_ax(
                    self.fig, ax,
                    xaxis_ypos=-0.1,
                    x_offset=[0.05, -0.15]
                )
                ax = change_box_to_only_y_arrow_ax(
                    self.fig, ax,
                    yaxis_xpos=0.0,
                    y_offset=[0.1, 0],
                    color=muscle_length_color
                )
                ax.tick_params(axis='x', direction='in', length=4, width=1, pad=0)
                ax.tick_params(
                    axis='y', direction='in', length=4, width=1, pad=1,
                    colors=muscle_length_color
                )
                if muscle_group_yticklabels_showflag['length'][j] is not True:
                    ax.set_yticklabels([])
                axes_muscle_length[-1].append(ax)

                ax = ax.twinx()
                ax.set_ylim([-0.1, 1.1])
                ax = change_box_to_only_y_arrow_ax(
                    self.fig, ax,
                    yaxis_xpos=1.0,
                    y_offset=[0.05, 0],
                    color=muscle_force_weight_color
                )
                ax.tick_params(
                    axis='y', direction='in', length=0, width=0, pad=-8,
                    labelcolor=muscle_force_weight_color
                )
                add_y_ticks(
                    ax=ax, yticks=[0, 0.5, 1], ticks_xpos=1,
                    length=0.04, linewidth=1,
                    color=muscle_force_weight_color
                )
                if muscle_group_yticklabels_showflag['force_weight'][j] is not True:
                    ax.set_yticklabels([])
                axes_muscle_force_weight[-1].append(ax)
            
            ax = self.fig.add_subplot(
                self.gs[
                    axes_muscle_indices_i['muscle'][0],
                    axes_muscle_indices_i['muscle'][1]
                ]
            )
            
            ax_muscle_offset = axes_muscle_info['axes_muscle_offset'][i]
            for spine in ax.spines.values():
                spine.set_visible(False)
            ax.get_xaxis().set_visible(False)
            ax.set_yticks([])
            ax.set_yticklabels([])
            ax.patch.set_facecolor('grey')
            ax.patch.set_alpha(0.3)
            bbox = ax.get_position()
            bbox.x0 += ax_muscle_offset[0][0]
            bbox.y0 += ax_muscle_offset[0][1]
            bbox.x1 += ax_muscle_offset[1][0]
            bbox.y1 += ax_muscle_offset[1][1]
            ax.set_position(bbox)
            axes_muscle.append(ax)

        return dict(
            muscle=axes_muscle,
            activation=axes_muscle_activation,
            length=axes_muscle_length,
            force_weight=axes_muscle_force_weight
        )

    @classmethod
    def plot_muscle_activation(cls, ax, s, activation, **kwargs):
        """plot_muscle_activation.

        Parameters
        ----------
        ax :
        s :
        activation :
        """
        alpha = kwargs.get("alpha", 1)
        if kwargs.get('fill', True):
            ax.fill_between(
                s, activation,
                color=kwargs["color"],
                alpha=alpha,
            )
        else:
            ax.plot(
                s, activation,
                color=kwargs["color"],
                alpha=alpha,
            )
        return ax

    @classmethod
    def plot_muscle_length(cls, ax, s, muscle_length, **kwargs):
        ax.plot(
            s, muscle_length,
            color=kwargs.get("color", muscle_length_color)
        )
        return ax

    @classmethod
    def plot_force_weight(cls, ax, s, force_weight, **kwargs):
        ax.plot(
            s, force_weight,
            color=kwargs.get("color", muscle_force_weight_color)
        )
        return ax


class TransverseMuscleFrame(MuscleFrameBase):
    """TransverseMuscleFrame.
    """

    def __init__(self, file_dict, fig_dict, gs_dict, **kwargs):
        """__init__.

        Parameters
        ----------
        file_dict :
        fig_dict :
        gs_dict :
        """
        MuscleFrameBase.__init__(
            self,
            file_dict=file_dict,
            fig_dict=fig_dict,
            gs_dict=gs_dict,
            **kwargs
        )

        self.axes_TM_info = kwargs["axes_TM_info"]
    
    def reset(self,):
        """reset.
        """
        self.axes_TM = MuscleFrameBase.reset(self, self.axes_TM_info)

class LongitudinalMuscleFrame(MuscleFrameBase):
    """LongitudinalMuscleFrame.
    """

    def __init__(self, file_dict, fig_dict, gs_dict, **kwargs):
        """__init__.

        Parameters
        ----------
        file_dict :
        fig_dict :
        gs_dict :
        """
        MuscleFrameBase.__init__(
            self,
            file_dict=file_dict,
            fig_dict=fig_dict,
            gs_dict=gs_dict,
            **kwargs
        )

        self.axes_LM_info = kwargs["axes_LM_info"]
    
    def reset(self,):
        """reset.
        """
        self.axes_LM = MuscleFrameBase.reset(self, self.axes_LM_info)

class ObliqueMuscleFrame(MuscleFrameBase):
    """ObliqueMuscleFrame.
    """

    def __init__(self, file_dict, fig_dict, gs_dict, **kwargs):
        MuscleFrameBase.__init__(
            self,
            file_dict=file_dict,
            fig_dict=fig_dict,
            gs_dict=gs_dict,
            **kwargs
        )

        self.axes_OM_info = kwargs["axes_OM_info"]
    
    def reset(self,):
        """__init__.

        Parameters
        ----------
        file_dict :
        fig_dict :
        gs_dict :
        """
        self.axes_OM = MuscleFrameBase.reset(self, self.axes_OM_info)



__doc__ = """
Forward Backward algorithm module.
"""

from tqdm import tqdm
import numpy as np

from coomm.algorithms.algorithm import Algorithm

class ForwardBackward(Algorithm):
    """ForwardBackward.
    """

    def __init__(self, rod, algo_config, **kwargs):
        """__init__.

        Parameters
        ----------
        rod :
        algo_config :
        """
        Algorithm.__init__(self, rod, algo_config)
        self.costate = Costate(self.static_rod.n_elements)
        self.stepsize = self.config.get('stepsize', 1e-8)
        self.iteration = 0
        self.done = False

        self.objects = kwargs.get('objects', kwargs.get('object', None))

    def update(self, iteration):
        """update.

        Parameters
        ----------
        iteration :
        """
        return iteration+1

    def run(self, max_iter_number=100_000, **kwargs):
        """run.

        Parameters
        ----------
        max_iter_number :
        kwargs :
        """
        print("Running the algorithm with objects:", self.objects)
        for _ in tqdm(range(max_iter_number)):
            self.iteration = self.update(self.iteration)
            if self.done:
                print("Finishing the algorithm at iternation", self.iteration)
                break
        print("Finishing the algorithm at maximum iternation", self.iteration)
        return

class Costate:
    """Costate.
    """

    def __init__(self, n_elements):
        """__init__.

        Parameters
        ----------
        n_elements :
        """
        # material frame
        self.internal_force = np.zeros((3, n_elements))
        self.internal_couple = np.zeros((3, n_elements-1))

        # lab frame
        self.internal_force_discrete_jump = np.zeros((3, n_elements))
        self.internal_couple_discrete_jump = np.zeros((3, n_elements))
        self.internal_force_derivative = np.zeros((3, n_elements))
        self.internal_couple_derivative = np.zeros((3, n_elements))



__doc__ = """
Processing scripts for the rendering
"""

import os, shutil
import numpy as np
from numba import njit

@njit(cache=True)
def process_position(position, offset, rotation):
    blocksize = position.shape[1]
    output_position = np.zeros((3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            for j in range(3):
                output_position[i, n] += (
                    rotation[i, j] * (position[j, n] - offset[j])
                )
    return output_position

@njit(cache=True)
def process_director(director, rotation):
    blocksize = director.shape[2]
    output_director = np.zeros((3, 3, blocksize))
    for n in range(blocksize):
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    output_director[i, j, n] += (
                        rotation[i, k] * director[k, j, n]
                    )
    return output_director

def check_folder(folder_name):
    if not (folder_name is None):
        if os.path.exists(folder_name):
            print('Clean up files in: {}/'.format(folder_name))
            shutil.rmtree(folder_name)
        print('Create the directory: {}/'.format(folder_name))
        os.mkdir(folder_name)



import numpy as np
import math

time_wave = np.sin(np.linspace(0, 100000, 1000000)) * 0.0


def draw_sucker(time_step, povray_data_folder, position, director, radius):
    with open(povray_data_folder + '/frame_sucker%04d.inc' % time_step, 'w') as file_inc:

        string = ""

        # import matplotlib.pyplot as plt
        # fig = plt.figure()
        # ax = fig.add_subplot(projection='3d')
        # ax.plot(position[0, :], position[1, :], position[2, :], 'k:')
        XX_list = []

        for j in [1,-1]:
            XX_side = []
            for k in [10,20,30, 40,50, 60,65,70,75, 80,85,90,92,94,96,98]:
                Y = 0.5*radius[k] * np.sin(np.linspace(0, 380, 10) / 180 * np.pi)
                Z = 0.5*radius[k] * np.cos(np.linspace(0, 380, 10) / 180 * np.pi)
                X = np.zeros_like(Y)
                circle = [X, Z, Y]
                # ax.plot(circle[0], circle[1], circle[2], 'g')
                a = np.dot(director[:, :, k].T, circle)

                ah = (45 * j) / 180 * np.pi
                Vh = director[2, :, k]
                surface_direction = director[0, :, k] * radius[k]
                # ax.plot([position[0, k], position[0, k] + 0.01 * surface_direction[0]],
                #         [position[1, k], position[1, k] + 0.01 * surface_direction[1]],
                #         [position[2, k], position[2, k] + 0.01 * surface_direction[2]], 'r')
                # ax.plot([position[0, k], position[0, k] + 0.01 * Vh[0]],
                #         [position[1, k], position[1, k] + 0.01 * Vh[1]],
                #         [position[2, k], position[2, k] + 0.01 * Vh[2]], 'b')

                R = rotation_matrix(Vh, ah)
                XX = np.dot(R, surface_direction)
                XX = XX + position[:, k]
                # ax.plot(XX[0], XX[1], XX[2], 'ro')
                XX_circle = XX[:, np.newaxis] + np.dot(R, a)

                # ax.plot(XX_circle[0], XX_circle[1], XX_circle[2], 'go')

                XX_side.append(XX_circle.T)
            XX_list.append(XX_side)
        # ax.set_xlim(-0.0, 0.2)
        # ax.set_ylim(-0.1, 0.1)
        # ax.set_zlim(-0.1, 0.1)
        # plt.show()
        # exit()
        print(np.array(XX_list).shape)
        # exit()
        for j in [0,1]:
            sucker_radius = np.linspace(0.003, 0.0005,len(XX_list[j]))
            for body_part_index in range(len(XX_list[j])):
                body_part = XX_list[j][body_part_index]
                n_elem = len(body_part)
                string += "sphere_sweep{\n\tb_spline %d" % n_elem

                for n in range(n_elem):
                    elem = body_part[n]
                    fake_shift = time_wave[time_step] * n / n_elem * 0.025
                    string += (",\n\t<%f,%f,%f>,%f" % (elem[0] ,#+ fake_shift * np.sin(time_step / 100),
                                                       elem[1] ,#+ fake_shift * np.cos(time_step / 100),
                                                       elem[2] ,#+ fake_shift * np.sin(time_step / 100) * np.cos(n),
                                                       sucker_radius[body_part_index]))
                string += "\n\ttexture{\n"
                string += "\t\tpigment{ color Yellow transmit %f }\n" % 0.0 #rgb < 0.45, 0.39, 1 >
                string += "\t\tfinish{ phong 1 }\n\t}\n"
                string += "\tscale<16,16,16>\n}\n"

        file_inc.writelines(
            string
        )


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    # fig = plt.figure()
    # ax = fig.add_subplot(projection='3d')

    # xh=np.linspace(0.15,0.02,200)
    # yh=np.linspace(0,1,200)
    # zh=((-6*yh+2)**3/3+(-6*yh+2)**2+4)/11
    #
    # ax.plot(xh, yh, zh,'k')
    # ax.plot(np.zeros_like(xh), yh, zh,'r')
    #
    #
    # for j in range(1,3):
    #     for i in range(0,200,20):
    #         XX_list = []
    #         for k in range(1,365,30):
    #             ah=(60*j)/180*np.pi
    #             V=[0,yh[i+1]-yh[i],zh[i+1]-zh[i]]
    #             Vh=V/np.linalg.norm(V)
    #
    #             X=xh[i]
    #             Y=0.25*xh[i]*np.sin(k/180*np.pi)
    #             Z=0.25*xh[i]*np.cos(k/180*np.pi)
    #
    #             R=rotation_matrix(Vh,ah)
    #             XX=np.dot(R, [X,Y,Z])
    #             XX = XX+ np.array([0,yh[i], zh[i]])
    #             XX_list.append(XX)
    #         data=np.array(XX_list)
    #         ax.plot(data[:,0], data[:,1], data[:,2])

    position = np.load("position.npy")
    director = np.load("director.npy")
    radius = np.load("radius.npy")
    print(radius)
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.plot(position[0, :], position[1, :], position[2, :], 'k:')
    XX_list = []
    for j in [1, -1]:
        for k in [20]:
            Y = 0.01 * np.sin(np.linspace(0, 360, 10) / 180 * np.pi)
            Z = 0.01 * np.cos(np.linspace(0, 360, 10) / 180 * np.pi)
            X = np.zeros_like(Y)
            circle = [X, Z, Y]
            ax.plot(circle[0], circle[1], circle[2], 'g')
            a = np.dot(director[:, :, k].T, circle)

            ah = (20 * j) / 180 * np.pi
            Vh = director[2, :, k]
            surface_direction = director[0, :, k] * radius[k] * 10
            ax.plot([position[0, k], position[0, k] + 0.01 * surface_direction[0]],
                    [position[1, k], position[1, k] + 0.01 * surface_direction[1]],
                    [position[2, k], position[2, k] + 0.01 * surface_direction[2]], 'r')
            ax.plot([position[0, k], position[0, k] + 0.01 * Vh[0]],
                    [position[1, k], position[1, k] + 0.01 * Vh[1]],
                    [position[2, k], position[2, k] + 0.01 * Vh[2]], 'b')

            R = rotation_matrix(Vh, ah)
            XX = np.dot(R, surface_direction)
            XX = XX + position[:, k]
            ax.plot(XX[0], XX[1], XX[2], 'ro')
            XX_circle = XX[:, np.newaxis] + np.dot(R, a)

            ax.plot(XX_circle[0], XX_circle[1], XX_circle[2], 'go')
            XX_list.append(XX)
        data = np.array(XX_list)
        # ax.plot(data[:, 0], data[:, 1], data[:, 2])
    ax.set_xlim(-0.1, 0.25)
    ax.set_ylim(-0.125, 0.125)
    ax.set_zlim(-0.125, 0.125)
    plt.show()
    print("done")



__doc__ = """
Base algorithm class
"""

import numpy as np
from coomm._rod_tool import StaticRod

class Algorithm:  # TODO: maybe rename the class??
    """Algorithm.
    """

    def __init__(self, rod, algo_config):
        """__init__.

        Parameters
        ----------
        rod :
        algo_config :
        """

        self.static_rod = StaticRod.get_rod(rod)
        self.config = algo_config

        self.ds = self.static_rod.rest_lengths / np.sum(self.static_rod.rest_lengths)
        self.s = np.insert(np.cumsum(self.ds), 0, 0)
        self.s_position = self.s.copy()
        self.s_director = (self.s[:-1] + self.s[1:])/2
        self.s_sigma = (self.s[:-1] + self.s[1:])/2
        self.s_kappa = self.s[1:-1]

    def run(self, **kwargs):
        """run.

        Parameters
        ----------
        kwargs :
        """
        raise NotImplementedError



__doc__ = """
Forward Backward Muscle model implementation.
"""

from os import stat
import numpy as np
from numba import njit

from tqdm import tqdm

from elastica._linalg import _batch_matvec, _batch_cross
from elastica._calculus import quadrature_kernel
from elastica.external_forces import inplace_addition

from coomm.algorithms.forward_backward import ForwardBackward
from coomm._rod_tool import (
    inverse,
    _lab_to_material,
    _material_to_lab,
    sigma_to_shear,
    average2D,
    calculate_dilatation
)

class ForwardBackwardMuscle(ForwardBackward):
    """ForwardBackwardMuscle.
    """

    def __init__(self, rod, muscles, algo_config, **kwargs):
        """__init__.

        Parameters
        ----------
        rod :
        muscles :
        algo_config :
        """
        ForwardBackward.__init__(self, rod, algo_config, **kwargs)
        self.activation_diff_tolerance = self.config['activation_diff_tolerance']
        self.muscles = muscles
        self.s_activations = []
        self.activations = []
        self.prev_activations = []
        for muscle in self.muscles:
            self.s_activations.append(muscle.s_activation.copy())
            self.activations.append(muscle.activation.copy())
            self.prev_activations.append(np.full(muscle.activation.shape, np.inf))

    def save_to_prev_activations(self, activations):
        """save_to_prev_activations.

        Parameters
        ----------
        activations :
        """
        for i in range(len(self.prev_activations)):
            self.prev_activations[i] = activations[i].copy()

    def update(self, iteration):
        """update.

        Parameters
        ----------
        iteration :

        Returns
        -------
        """
        self.save_to_prev_activations(self.activations)

        # find the equlibrium for the current muscle activations
        self.find_equilibrium_strain(
            self.static_rod.sigma, self.static_rod.kappa,
            self.static_rod.shear_matrix, self.static_rod.bend_matrix,
            self.static_rod.dilatation, self.static_rod.voronoi_dilatation,
            *self.calculate_total_muscle_forces_couples()
        )

        # forward path
        self.static_rod.update_from_strain(
            self.static_rod.sigma, self.static_rod.kappa
        )

        # update cost-related terms in objects
        self.objects(
            position=self.static_rod.position_collection,
            director=self.static_rod.director_collection,
            radius=self.static_rod.radius
        )

        # backward path
        self.discrete_cost_gradient_condition()
        self.continuous_cost_gradient_condition()

        self.costate_backward_evolution(
            self.static_rod.rest_lengths, 
            self.static_rod.director_collection, 
            self.static_rod.sigma,
            self.costate.internal_force_discrete_jump,
            self.costate.internal_couple_discrete_jump,
            self.costate.internal_force_derivative,
            self.costate.internal_couple_derivative,
            self.costate.internal_force, self.costate.internal_couple
        )
        
        # update activations
        self.update_activations(
            self.find_target_activations()
        )

        # check if the updated activations are similar with previous ones
        self.done = self.check_activations_difference()

        return ForwardBackward.update(self, iteration)

    def calculate_total_muscle_forces_couples(self):
        """calculate_total_muscle_forces_couples.

        Returns
        -------
        muscle_forces:
        muscle_couples:
        """

        muscle_forces = np.zeros(self.static_rod.sigma.shape)
        muscle_couples = np.zeros(self.static_rod.kappa.shape)

        for muscle, activation in zip(self.muscles, self.activations):
            muscle.apply_activation(activation)
            muscle(self.static_rod)
            inplace_addition(muscle_forces, muscle.internal_force)
            inplace_addition(muscle_couples, muscle.internal_couple)

        return muscle_forces, muscle_couples

    @staticmethod
    @njit(cache=True)
    def find_equilibrium_strain(
        sigma, kappa,
        shear_matrix, bend_matrix,
        dilatation, voronoi_dilatation,
        muscle_forces, muscle_couples
    ):
        kappa[:, :] = - _batch_matvec(
            inverse(bend_matrix/voronoi_dilatation**3), muscle_couples
        )
        sigma[:, :] = - _batch_matvec(
            inverse(shear_matrix/dilatation), muscle_forces
        )

    def discrete_cost_gradient_condition(self,):
        self.costate.internal_force_discrete_jump[:, -1] = (
            - self.objects.cost_gradient.discrete.wrt_position[:, -1]
        )
        # print(self.costate.internal_force_discrete_jump[:, :])
        self.costate.internal_couple_discrete_jump[:, -1] = (
            - self.objects.cost_gradient.discrete.wrt_director[:, -1]
        )

    def continuous_cost_gradient_condition(self,):
        self.costate.internal_force_derivative[:, :] = (
            self.objects.cost_gradient.continuous.wrt_position
        )
        self.costate.internal_couple_derivative[:, :] = (
            self.objects.cost_gradient.continuous.wrt_director
        )

    # This method is not completed. It only compute tip jump rather than any jump at any location
    @staticmethod
    @njit(cache=True)
    def costate_backward_evolution(
        rest_lengths, director, sigma,
        internal_force_lab_frame_at_tip,            # This name needs to be changed to internal_force_lab_frame_jump or something similar
        internal_couple_lab_frame_at_tip,           # This name needs to be changed
        internal_force_lab_frame_derivative,
        internal_couple_lab_frame_derivative,
        internal_force, internal_couple
    ):
        blocksize = rest_lengths.shape[0]
        internal_force_lab_frame = np.zeros((3, blocksize))
        internal_couple_lab_frame = np.zeros((3, blocksize))
        shear = sigma_to_shear(sigma)
        dilatation, _ = calculate_dilatation(shear)

        # TODO: The next line is incomplete
        internal_force_lab_frame[:, -1] = internal_force_lab_frame_at_tip[:, -1]
        
        force_derivative = average2D(internal_force_lab_frame_derivative)

        # n_s = f
        for k in range(blocksize-1):
            internal_force_lab_frame[:, -1-k-1] = internal_force_lab_frame[:, -1-k] - (
                force_derivative[:, -1-k] * 0.5 * (rest_lengths[-1-k] * dilatation[-1-k] + rest_lengths[-1-k-1] * dilatation[-1-k-1])
            )
        internal_force[:, :] = _lab_to_material(director, internal_force_lab_frame)

        internal_couple_lab_frame[:, -1] = internal_couple_lab_frame_at_tip[:, -1]

        # m_s = -r_s x n + c
        internal_couple_lab_frame_derivative[:, :] -= _batch_cross(
            _material_to_lab(director, shear), internal_force_lab_frame
        )
        couple_derivative = average2D(internal_couple_lab_frame_derivative)
        for k in range(blocksize-1):
            internal_couple_lab_frame[:, -1-k-1] = internal_couple_lab_frame[:, -1-k] - (
                couple_derivative[:, -1-k] * 0.5 * (rest_lengths[-1-k] * dilatation[-1-k] + rest_lengths[-1-k-1] * dilatation[-1-k-1])
            )
        internal_couple[:, :] = average2D(
            _lab_to_material(director, internal_couple_lab_frame)
        )

    def find_target_activations(self):
        """find_target_activations.
        """
        target_activations = []
        for muscle in self.muscles:
            muscle.apply_activation(np.ones(muscle.activation.shape))
            muscle(self.static_rod)
            target_activations.append(
                self.calculate_target_activation(
                    self.costate.internal_force,
                    self.costate.internal_couple,
                    self.static_rod.dilatation,
                    self.static_rod.voronoi_dilatation,
                    self.static_rod.shear_matrix,
                    self.static_rod.bend_matrix,
                    muscle.internal_force, muscle.internal_couple
                )
            )
        return target_activations

    @staticmethod
    @njit(cache=True)
    def calculate_target_activation(
        internal_force, internal_couple,
        dilatation, voronoi_dilatation,
        shear_matrix, bend_matrix,
        muscle_internal_force, muscle_internal_couple
    ):
        blocksize = internal_force.shape[1]
        target_activation = np.zeros(blocksize)
        temp_shear = _batch_matvec(
            inverse(shear_matrix/dilatation),
            muscle_internal_force
        )
        temp_kappa = _batch_matvec(
            inverse(bend_matrix/voronoi_dilatation),
            muscle_internal_couple
        )
        temp_force_innerproduct = np.zeros(blocksize)
        temp_couple_innerproduct = np.zeros(blocksize+1)
        # for k in range(blocksize-1):
        #     for i in range(3):
        #         temp_force_innerproduct[k] += internal_force[i, k] * temp_shear[i, k]
        #         temp_couple_innerproduct[k+1] += internal_couple[i, k+1] * temp_kappa[i, k+1]
        # temp_force_innerproduct[-1] += internal_force[i, -1] * temp_shear[i, -1]

        for k in range(blocksize):
            for i in range(3):
                temp_force_innerproduct[k] += internal_force[i, k] * temp_shear[i, k]
        for k in range(blocksize-1):
            for i in range(3):
                temp_couple_innerproduct[k+1] += internal_couple[i, k] * temp_kappa[i, k]
        temp_couple_innerproduct[0] = 2*temp_couple_innerproduct[1]-temp_couple_innerproduct[2]
        temp_couple_innerproduct[-1] = 2*temp_couple_innerproduct[-2]-temp_couple_innerproduct[-3]

        for k in range(blocksize):
            target_activation[k] = -temp_force_innerproduct[k] - 0.5*(
                temp_couple_innerproduct[k] + temp_couple_innerproduct[k+1]
            )
        return target_activation

    def update_activations(self, target_activations):
        """update_activations.

        Parameters
        ----------
        target_activations :
        """
        for activation, target_activation, in zip(self.activations, target_activations):
            activation[:] -= self.stepsize * (activation - target_activation)
            activation[:] = np.clip(activation, 0, 1)

    def check_activations_difference(self):
        """check_activations_difference

        Returns
        -------
        """
        norm = 0
        for prev_activation, activation in zip(self.prev_activations, self.activations):
            norm += np.sum((activation-prev_activation)**2)/activation.shape[0]
        norm /= len(self.activations)
        return True if norm < self.activation_diff_tolerance else False
